# -*- coding: utf-8 -*-
"""
Functions for fetching annotations (from the internet, if necessary)
"""

from collections import defaultdict
from pathlib import Path
import re
import shutil

from nilearn.datasets.utils import _fetch_file

from neuromaps.datasets.utils import (get_data_dir, get_dataset_info,
                                      _get_token, _get_session)

MATCH = re.compile(
    r'source-(\S+)_desc-(\S+)_space-(\S+)_(?:den|res)-(\d+[k|m]{1,2})_'
)


def _groupby_match(fnames, return_single=False):
    """"
    Groups files in `fnames` by (source, desc, space, res/den)

    Parameters
    ----------
    fnames : list-of-str
        Filenames to be grouped
    return_single : bool, optional
        If there is only group of filenames return a list instead of a dict.
        Default: False

    Returns
    -------
    groups : dict-of-str
        Where keys are tuple (source, desc, space, res/den) and values are
        lists of filenames
    """

    out = defaultdict(list)
    for fn in fnames:
        out[MATCH.search(fn).groups()].append(fn)

    out = {k: v if len(v) > 1 else v[0] for k, v in out.items()}

    if return_single and len(out) == 1:
        out = list(out.values())[0]

    return out


def _match_annot(info, **kwargs):
    """
    Matches datasets in `info` to relevant keys

    Parameters
    ----------
    info : list-of-dict
        Information on annotations
    kwargs : key-value pairs
        Values of data in `info` on which to match

    Returns
    -------
    matched : list-of-dict
        Annotations with specified values for keys
    """

    # tags should always be a list
    tags = kwargs.get('tags')
    if tags is not None and isinstance(tags, str):
        kwargs['tags'] = [tags]

    # 'den' and 'res' are a special case because these are mutually exclusive
    # values (only one will ever be set for a given annotation) so we want to
    # match on _either_, not both, if and only if both are provided as keys.
    # if only one is specified as a key then we should exclude the other!
    denres = []
    for vals in (kwargs.get('den'), kwargs.get('res')):
        vals = [vals] if isinstance(vals, str) else vals
        if vals is not None:
            denres.extend(vals)

    out = []
    for dset in info:
        match = True
        for key in ('source', 'desc', 'space', 'hemi', 'tags', 'format'):
            comp, value = dset.get(key), kwargs.get(key)
            if value is None:
                continue
            elif value is not None and comp is None:
                match = False
            elif isinstance(value, str):
                if value != 'all':
                    match = match and comp == value
            else:
                func = all if key == 'tags' else any
                match = match and func(f in comp for f in value)
        if len(denres) > 0:
            match = match and (dset.get('den') or dset.get('res')) in denres
        if match:
            out.append(dset)

    return out


def available_annotations(source=None, desc=None, space=None, den=None,
                          res=None, hemi=None, tags=None, format=None,
                          return_restricted=False):
    """
    Lists datasets available via :func:`~.fetch_annotation`

    Parameters
    ----------
    source, desc, space, den, res, hemi, tags, format : str or list-of-str
        Values on which to match annotations. If not specified annotations with
        any value for the relevant key will be matched. Default: None
    return_restricted : bool, optional
        Whether to return restricted annotations. These will only be accesible
        with a valid OSF token. Default: True

    Returns
    -------
    datasets : list-of-str
        List of available annotations
    """

    info = _match_annot(get_dataset_info('annotations', return_restricted),
                        source=source, desc=desc, space=space, den=den,
                        res=res, hemi=hemi, tags=tags, format=format)
    fnames = [dset['fname'] for dset in info]

    return list(_groupby_match(fnames, return_single=False).keys())


def available_tags(return_restricted=False):
    """
    Returns available tags for querying annotations

    Parameters
    ----------
    return_restricted : bool, optional
        Whether to return restricted annotations. These will only be accesible
        with a valid OSF token. Default: True


    Returns
    -------
    tags : list-of-str
        Available tags
    """

    tags = set()
    for dset in get_dataset_info('annotations', return_restricted):
        if dset['tags'] is not None:
            tags.update(dset['tags'])
    return sorted(tags)


def fetch_annotation(*, source=None, desc=None, space=None, den=None, res=None,
                     hemi=None, tags=None, format=None, return_single=True,
                     token=None, data_dir=None, verbose=1):
    """
    Downloads files for brain annotations matching requested variables

    Parameters
    ----------
    source, desc, space, den, res, hemi, tags, format : str or list-of-str
        Values on which to match annotations. If not specified annotations with
        any value for the relevant key will be matched. Default: None
    return_single : bool, optional
        If only one annotation is found matching input parameters return the
        list of filepaths instead of the standard dictionary. Default: True
    token : str, optional
        OSF personal access token for accessing restricted annotations. Will
        also check the environmental variable 'NEUROMAPS_OSF_TOKEN' if not
        provided; if that is not set no token will be provided and restricted
        annotations will be inaccessible. Default: None
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NEUROMAPS_DATA'; if that is not set, will
        use `~/neuromaps-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Returns
    -------
    data : dict
        Dictionary of downloaded annotations where dictionary keys are tuples
        (source, desc, space, den/res) and values are lists of corresponding
        filenames
    """

    # check input parameters to ensure we're fetching _something_
    supplied = False
    for val in (source, desc, space, den, res, hemi, tags, format):
        if val is not None:
            supplied = True
            break
    if not supplied:
        raise ValueError('Must provide at least one parameters on which to '
                         'match annotations. If you want to fetch all '
                         'annotations set any of the parameters to "all".')

    # get info on datasets we need to fetch
    token = _get_token(token=token)
    return_restricted = False if (token is None or not token) else True
    data_dir = get_data_dir(data_dir=data_dir)
    info = _match_annot(get_dataset_info('annotations', return_restricted),
                        source=source, desc=desc, space=space, den=den,
                        res=res, hemi=hemi, tags=tags, format=format)
    if verbose > 1:
        print(f'Identified {len(info)} datsets matching specified parameters')

    # get session for requests
    session = _get_session(token=token)

    # TODO: current work-around to handle that _fetch_files() does not support
    # session instances. hopefully a future version will and we can just use
    # that function to handle this instead of calling _fetch_file() directly
    data = []
    for dset in info:
        fn = Path(data_dir) / 'annotations' / dset['rel_path'] / dset['fname']
        if not fn.exists():
            dl_file = _fetch_file(dset['url'], str(fn.parent), verbose=verbose,
                                  md5sum=dset['checksum'], session=session)
            shutil.move(dl_file, fn)
        data.append(str(fn))

    return _groupby_match(data, return_single=return_single)



# -*- coding: utf-8 -*-
"""
For testing neuromaps.transforms functionality
"""

import pytest

from neuromaps import transforms


@pytest.mark.xfail
def test__regfusion_project():
    assert False


@pytest.mark.xfail
def test__vol_to_surf():
    assert False


@pytest.mark.xfail
def test_mni152_to_civet():
    assert False


@pytest.mark.xfail
def test_mni152_to_fsaverage():
    assert False


@pytest.mark.xfail
def test_mni152_to_fslr():
    assert False


@pytest.mark.xfail
def test_mni152_to_mni152():
    assert False


def test__check_hemi():
    d, h = zip(*transforms._check_hemi('test', 'L'))
    assert d == ('test',) and h == ('L',)

    for d, h in (('test', None), ('test', 'invalid_hemi')):
        with pytest.raises(ValueError):
            transforms._check_hemi(d, h)


@pytest.mark.xfail
def test__surf_to_surf():
    assert False


@pytest.mark.xfail
def test_civet_to_fslr():
    assert False


@pytest.mark.xfail
def test_fslr_to_civet():
    assert False


@pytest.mark.xfail
def test_civet_to_fsaverage():
    assert False


@pytest.mark.xfail
def test_fsaverage_to_civet():
    assert False


@pytest.mark.xfail
def test_fslr_to_fsaverage():
    assert False


@pytest.mark.xfail
def test_fsaverage_to_fslr():
    assert False


@pytest.mark.xfail
def test_civet_to_civet():
    assert False


@pytest.mark.xfail
def test_fslr_to_fslr():
    assert False


@pytest.mark.xfail
def test_fsaverage_to_fsaverage():
    assert False



# -*- coding: utf-8 -*-
"""
Functionality for transforming files between spaces
"""

import os
from pathlib import Path

import nibabel as nib
from nilearn import image as nimage
import numpy as np
from scipy.interpolate import interpn

from neuromaps.datasets import (ALIAS, DENSITIES, fetch_atlas,
                                fetch_regfusion, get_atlas_dir)
from neuromaps.images import (construct_shape_gii, load_gifti, load_nifti,
                              load_data)
from neuromaps.utils import tmpname, run

METRICRESAMPLE = 'wb_command -metric-resample {metric} {src} {trg} ' \
                 'ADAP_BARY_AREA {out} -area-metrics {srcarea} {trgarea} ' \
                 '-current-roi {srcmask}'
LABELRESAMPLE = 'wb_command -label-resample {metric} {src} {trg} ' \
                'ADAP_BARY_AREA {out} -area-metrics {srcarea} {trgarea} ' \
                '-current-roi {srcmask}'
MASKSURF = 'wb_command -metric-mask {out} {trgmask} {out}'
SURFFMT = 'tpl-{space}{trg}_den-{den}_hemi-{hemi}_sphere.surf.gii'
VAFMT = 'tpl-{space}_den-{den}_hemi-{hemi}_desc-vaavg_midthickness.shape.gii'
MLFMT = 'tpl-{space}_den-{den}_hemi-{hemi}_desc-nomedialwall_dparc.label.gii'
DENSITY_MAP = {
    642: '1k', 2562: '3k', 4002: '4k', 7842: '8k', 10242: '10k',
    32492: '32k', 40962: '41k', 163842: '164k'
}


def _estimate_density(data, hemi=None):
    """
    Tries to estimate standard density of `data`

    Parameters
    ----------
    data : (2,) tuple of (2,) tuple of str or os.PathLike or nib.GiftiImage
        Input data for (src, trg), where src and trg are len-2 tuples of images
    hemi : {'L', 'R'}, optional
        If entries of `data` are not tuples this specifies the hemisphere the
        data represent. Default: None

    Returns
    -------
    density : tuple-of-str
        Tuple of strings representing approximate density of data

    Raises
    ------
    ValueError
        If density of `data` is not one of the standard expected values
    """

    densities = tuple()
    for img in data:
        # if `img` is actually just a density string then return that!
        if img in DENSITY_MAP.values():
            densities += (img,)
            continue

        # we expect `img` here to actually be a tuple-of-str (L/R hemisphere)
        # if not, hemi is required
        img, hemi = zip(*_check_hemi(img, hemi))

        # confirm that both entries in `img` (if a tuple) have same density
        n_vert = [len(load_data(d)) for d in img]
        if not all(n == n_vert[0] for n in n_vert):
            raise ValueError('Provided data have different resolutions across '
                             'hemispheres?')
        else:
            n_vert = n_vert[0]

        # get string-abbreviated density for data
        density = DENSITY_MAP.get(n_vert)
        if density is None:
            raise ValueError('Provided data resolution is non-standard. '
                             f'Number of vertices estimated in data: {n_vert}')
        densities += (density,)

    return densities


def _regfusion_project(data, ras, affine, method='linear'):
    """
    Project `data` to `ras` space using regfusion

    Parameters
    ----------
    data : (X, Y, Z[, V]) array_like
        Input (volumetric) data to be projected to the surface
    ras : (N, 3) array_like
        Coordinates of surface points derived from registration fusion
    affine (4, 4) array_like
        Affine mapping `data` to `ras`-space coordinates
    method : {'nearest', 'linear'}, optional
        Method for projection. Default: 'linear'

    Returns
    -------
    projected : (N, V) array_like
        Input `data` projected to the surface
    """

    data, ras, affine = np.asarray(data), np.asarray(ras), np.asarray(affine)
    coords = nib.affines.apply_affine(np.linalg.inv(affine), ras)
    volgrid = [range(data.shape[i]) for i in range(3)]
    if data.ndim == 3:
        projected = interpn(volgrid, data, coords, method=method)
    elif data.ndim == 4:
        projected = np.column_stack([
            interpn(volgrid, data[..., n], coords, method=method)
            for n in range(data.shape[-1])
        ])

    return construct_shape_gii(projected.squeeze())


def _vol_to_surf(img, space, density, method='linear'):
    """
    Projects `img` to the surface defined by `space` and `density`

    Parameters
    ----------
    img : niimg_like, str, or os.PathLike
        Image to be projected to the surface
    den : str
        Density of desired output space
    space : str
        Desired output space
    method : {'nearest', 'linear'}, optional
        Method for projection. Default: 'linear'

    Returns
    -------
    projected : (2,) tuple-of-nib.GiftiImage
        Left [0] and right [1] hemisphere projected `image` data
    """

    space = ALIAS.get(space, space)
    if space not in DENSITIES:
        raise ValueError(f'Invalid space argument: {space}')
    if density not in DENSITIES[space]:
        raise ValueError(f'Invalid density for {space} space: {density}')
    if method not in ('nearest', 'linear'):
        raise ValueError('Invalid method argument: {method}')

    img = load_nifti(img)
    out = ()
    for ras in fetch_regfusion(space)[density]:
        out += (_regfusion_project(img.get_fdata(), np.loadtxt(ras),
                                   img.affine, method=method),)

    return out


def mni152_to_civet(img, civet_density='41k', method='linear'):
    """
    Projects `img` in MNI152 space to CIVET surface

    Parameters
    ----------
    img : str or os.PathLike or niimg_like
        Image in MNI152 space to be projected
    civet_density : {'41k'}, optional
        Desired output density of CIVET surface. Default: '41k'
    method : {'nearest', 'linear'}, optional
        Method for projection. Specify 'nearest' if `img` is a label image.
        Default: 'linear'

    Returns
    -------
    civet : (2,) tuple-of-nib.GiftiImage
        Projected `img` on CIVET surface
    """

    if civet_density == '164k':
        raise NotImplementedError('Cannot perform registration fusion to '
                                  'CIVET 164k space yet.')

    return _vol_to_surf(img, 'civet', civet_density, method)


def mni152_to_fsaverage(img, fsavg_density='41k', method='linear'):
    """
    Projects `img` in MNI152 space to fsaverage surface

    Parameters
    ----------
    img : str or os.PathLike or niimg_like
        Image in MNI152 space to be projected
    fsavg_density : {'3k', '10k', '41k', '164k'}, optional
        Desired output density of fsaverage surface. Default: '41k'
    method : {'nearest', 'linear'}, optional
        Method for projection. Specify 'nearest' if `img` is a label image.
        Default: 'linear'

    Returns
    -------
    fsaverage : (2,) tuple-of-nib.GiftiImage
        Projected `img` on fsaverage surface
    """

    return _vol_to_surf(img, 'fsaverage', fsavg_density, method)


def mni152_to_fslr(img, fslr_density='32k', method='linear'):
    """
    Projects `img` in MNI152 space to fsLR surface

    Parameters
    ----------
    img : str or os.PathLike or niimg_like
        Image in MNI152 space to be projected
    fslr_density : {'32k', '164k'}, optional
        Desired output density of fsLR surface. Default: '32k'
    method : {'nearest', 'linear'}, optional
        Method for projection. Specify 'nearest' if `img` is a label image.
        Default: 'linear'

    Returns
    -------
    fsLR : (2,) tuple-of-nib.GiftiImage
        Projected `img` on fsLR surface
    """

    if fslr_density in ('4k', '8k'):
        raise NotImplementedError('Cannot perform registration fusion to '
                                  f'fsLR {fslr_density} space yet.')

    return _vol_to_surf(img, 'fsLR', fslr_density, method)


def mni152_to_mni152(img, target='1mm', method='linear'):
    """
    Resamples `img` to `target` image (if supplied) or target `resolution`

    Parameters
    ----------
    img : str or os.PathLike or niimg_like
        Image in MNI152 space to be resampled
    target : {str, os.PathLike, niimg_like} or {'1mm', '2mm', '3mm'}, optional
        Image in MNI152 space to which `img` should be resampled. Can
        alternatively specify desired resolution of output resample image.
        Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `img` is a label image.
        Default: 'linear'

    Returns
    -------
    resampled : nib.Nifti1Image
        Resampled input `img`
    """

    if target in DENSITIES['MNI152']:
        target = fetch_atlas('MNI152', target)['T1w']

    out = nimage.resample_to_img(img, target, interpolation=method)

    return out


def _check_hemi(data, hemi):
    """ Utility to check that `data` and `hemi` jibe

    Parameters
    ----------
    data : str or os.PathLike or tuple
        Input data
    hemi : str
        Hemisphere(s) corresponding to `data

    Returns
    -------
    zipped : zip
        Zipped instance of `data` and `hemi`
    """

    if isinstance(data, (str, os.PathLike)) or not hasattr(data, '__len__'):
        data = (data,)
    if len(data) == 1 and hemi is None:
        raise ValueError('Must specify `hemi` when only 1 data file supplied')
    if hemi is not None and isinstance(hemi, str) and hemi not in ('L', 'R'):
        raise ValueError(f'Invalid hemisphere designation: {hemi}')
    elif hemi is not None and isinstance(hemi, str):
        hemi = (hemi,)
    elif hemi is not None and any(h not in ('L', 'R') for h in hemi):
        raise ValueError(f'Invalid hemisphere designations: {hemi}')
    elif hemi is None:
        hemi = ('L', 'R')

    return zip(data, hemi)


def _surf_to_surf(data, srcparams, trgparams, method='linear', hemi=None):
    """
    Resamples surface `data` to another surface

    Parameters
    ----------
    data : str or os.Pathlike or tuple
        Filepath(s) to data. If not a tuple then `hemi` must be specified. If
        a tuple then it is assumed that files are ('left', 'right')
    srcparams, trgparams : dict
        Dictionary with keys ['space', 'den', 'trg']
    method : {'nearest', 'linear'}, optional
        Method for resampling. Default: 'linear'
    hemi : str or None
        Hemisphere of `data` if `data` is a single image. Default: None

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    methods = ('nearest', 'linear')
    if method not in methods:
        raise ValueError(f'Invalid method: {method}. Must be one of {methods}')

    keys = ('space', 'den', 'trg')
    for key in keys:
        if key not in srcparams:
            raise KeyError(f'srcparams missing key: {key}')
        if key not in trgparams:
            raise KeyError(f'trgparams missing key: {key}')

    for val in (srcparams, trgparams):
        space, den = val['space'], val['den']
        if den not in DENSITIES[space]:
            raise ValueError(f'Invalid density for {space} space: {den}')

    # if our source and target are identical just return the loaded data
    if srcparams == trgparams:
        data, _ = zip(*_check_hemi(data, hemi))
        return tuple(load_gifti(d) for d in data)

    # get required atlas / templates for transforming between spaces
    for atl in (srcparams, trgparams):
        fetch_atlas(atl['space'], atl['den'])
    srcdir = get_atlas_dir(srcparams['space'])
    trgdir = get_atlas_dir(trgparams['space'])

    resampled = ()
    func = METRICRESAMPLE if method == 'linear' else LABELRESAMPLE
    for img, hemi in _check_hemi(data, hemi):
        srcparams['hemi'] = trgparams['hemi'] = hemi
        try:
            img = Path(img).resolve()
            tmpimg = None
        except TypeError:
            tmpimg = tmpname(suffix='.gii')
            nib.save(img, tmpimg)
            img = Path(tmpimg).resolve()
        params = dict(
            metric=img,
            out=tmpname('.func.gii'),
            src=srcdir / SURFFMT.format(**srcparams),
            trg=trgdir / SURFFMT.format(**trgparams),
            srcarea=srcdir / VAFMT.format(**srcparams),
            trgarea=trgdir / VAFMT.format(**trgparams),
            srcmask=srcdir / MLFMT.format(**srcparams),
            trgmask=trgdir / MLFMT.format(**trgparams)
        )
        for fn in (func, MASKSURF):
            run(fn.format(**params), quiet=True)
        resampled += (construct_shape_gii(load_data(params['out'])),)
        params['out'].unlink()
        if tmpimg is not None:
            tmpimg.unlink()

    return resampled


def civet_to_fslr(data, target_density='32k', hemi=None, method='linear'):
    """
    Resamples `data` on CIVET surface to the fsLR surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input CIVET data to be resampled to fsLR surface
    target_density : {'4k', '8k', '32k', '164k'}, optional
        Desired density of output fsLR surface. Default: '32k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='civet', den=density, trg='_space-fsLR')
    trgparams = dict(space='fsLR', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fslr_to_civet(data, target_density='41k', hemi=None, method='linear'):
    """
    Resamples `data` on fsLR surface to the CIVET surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsLR data to be resampled to CIVET surface
    target_density : {'41k', '164k'}, optional
        Desired density of output CIVET surface. Default: '41k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsLR', den=density, trg='')
    trgparams = dict(space='civet', den=target_density, trg='_space-fsLR')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def civet_to_fsaverage(data, target_density='41k', hemi=None, method='linear'):
    """
    Resamples `data` on CIVET surface to the fsaverage surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input CIVET data to be resampled to fsaverage surface
    target_density : {'3k', '10k', '41k', '164k'}, optional
        Desired density of output fsaverage surface. Default: '32k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='civet', den=density, trg='_space-fsaverage')
    trgparams = dict(space='fsaverage', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fsaverage_to_civet(data, target_density='41k', hemi=None, method='linear'):
    """
    Resamples `data` on fsaverage surface to the CIVET surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsaverage data to be resampled to CIVET surface
    target_density : {'41k', '164k'}, optional
        Desired density of output CIVET surface. Default: '41k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsaverage', den=density, trg='')
    trgparams = dict(space='civet', den=target_density, trg='_space-fsaverage')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fslr_to_fsaverage(data, target_density='41k', hemi=None, method='linear'):
    """
    Resamples `data` on fsLR surface to the fsaverage surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsLR data to be resampled to fsaverage surface
    target_density : {'3k', '10k', '41k', '164k'}, optional
        Desired density of output fsaverage surface. Default: '41k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsLR', den=density, trg='_space-fsaverage')
    trgparams = dict(space='fsaverage', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fsaverage_to_fslr(data, target_density='32k', hemi=None, method='linear'):
    """
    Resamples `data` on fsaverage surface to the fsLR surface

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsaverage data to be resampled to fsLR surface
    target_density : {'4k', '8k', '32k', '164k'}, optional
        Desired density of output fsLR surface. Default: '32k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsaverage', den=density, trg='')
    trgparams = dict(space='fsLR', den=target_density, trg='_space-fsaverage')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def civet_to_civet(data, target_density='41k', hemi=None, method='linear'):
    """
    Resamples `data` on CIVET surface to new density

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input CIVET data to be resampled
    target_density : {'41k', '164k'}, optional
        Desired density of output surface. Default: '41k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new surface
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='civet', den=density, trg='')
    trgparams = dict(space='civet', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fslr_to_fslr(data, target_density='32k', hemi=None, method='linear'):
    """
    Resamples `data` on fsLR surface to new density

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsLR data to be resampled
    target_density : {'4k', '8k', '32k', '164k'}, optional
        Desired density of output surface. Default: '32k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new density
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsLR', den=density, trg='')
    trgparams = dict(space='fsLR', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)


def fsaverage_to_fsaverage(data, target_density='41k', hemi=None,
                           method='linear'):
    """
    Resamples `data` on fsaverage surface to new density

    Parameters
    ----------
    data : str or os.PathLike or nib.GiftiImage or tuple
        Input fsaverage data to be resampled
    target_density : {'3k', '10k', '41k', '164k'}, optional
        Desired density of output surface. Default: '41k'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple this specifies the hemisphere the data are
        representing. Default: None
    method : {'nearest', 'linear'}, optional
        Method for resampling. Specify 'nearest' if `data` are label images.
        Default: 'linear'

    Returns
    -------
    resampled : tuple-of-nib.GiftiImage
        Input `data` resampled to new density
    """

    density, = _estimate_density((data,), hemi=hemi)
    srcparams = dict(space='fsaverage', den=density, trg='')
    trgparams = dict(space='fsaverage', den=target_density, trg='')
    return _surf_to_surf(data, srcparams, trgparams, method, hemi)



# -*- coding: utf-8 -*-
"""
Functionality for plotting
"""

from matplotlib import colors as mcolors, pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa
from nilearn.plotting import plot_surf
import numpy as np

from neuromaps.datasets import ALIAS, fetch_atlas
from neuromaps.images import load_gifti
from neuromaps.transforms import _check_hemi

HEMI = dict(L='left', R='right')
plt.cm.register_cmap(
    'caret_blueorange', mcolors.LinearSegmentedColormap.from_list('blend', [
        '#00d2ff', '#009eff', '#006cfe', '#0043fe',
        '#fd4604', '#fe6b01', '#ffd100', '#ffff04'
    ])
)


def plot_surf_template(data, template, density, surf='inflated',
                       hemi=None, data_dir=None, mask_medial=True, **kwargs):
    """
    Plots `data` on `template` surface

    Parameters
    ----------
    data : str or os.PathLike or tuple-of-str
        Path to data file(s) to be plotted. If tuple, assumes (left, right)
        hemisphere.
    template : {'civet', 'fsaverage', 'fsLR'}
        Template on which `data` is defined
    density : str
        Resolution of template
    surf : str, optional
        Surface on which `data` should be plotted. Must be valid for specified
        `space`. Default: 'inflated'
    hemi : {'L', 'R'}, optional
        If `data` is not a tuple, which hemisphere it should be plotted on.
        Default: None
    mask_medial : bool, optional
        Whether to mask vertices along the medial wall. Default: True
    kwargs : key-value pairs
        Passed directly to `nilearn.plotting.plot_surf`

    Returns
    -------
    fig : matplotlib.Figure instance
        Plotted figure
    """

    atlas = fetch_atlas(template, density, data_dir=data_dir, verbose=0)
    template = ALIAS.get(template, template)
    if template == 'MNI152':
        raise ValueError('Cannot plot MNI152 on the surface. Try performing '
                         'registration fusion to project data to the surface '
                         'and plotting the projection instead.')
    surf, medial = atlas[surf], atlas['medial']

    opts = dict(alpha=1.0, threshold=np.spacing(1))
    opts.update(**kwargs)
    if kwargs.get('bg_map') is not None and kwargs.get('alpha') is None:
        opts['alpha'] = 'auto'

    data, hemispheres = zip(*_check_hemi(data, hemi))
    n_surf = len(data)
    fig, axes = plt.subplots(n_surf, 2, subplot_kw={'projection': '3d'})
    axes = (axes,) if n_surf == 1 else axes.T
    for row, hemi, img in zip(axes, hemispheres, data):
        geom = load_gifti(getattr(surf, hemi)).agg_data()
        img = load_gifti(img).agg_data().astype('float32')
        # set medial wall to NaN; this will avoid it being plotted
        if mask_medial:
            med = load_gifti(getattr(medial, hemi)).agg_data().astype(bool)
            img[np.logical_not(med)] = np.nan

        for ax, view in zip(row, ['lateral', 'medial']):
            ax.disable_mouse_rotation()
            plot_surf(geom, img, hemi=HEMI[hemi], axes=ax, view=view, **opts)
            poly = ax.collections[0]
            poly.set_facecolors(
                _fix_facecolors(ax, poly._original_facecolor,
                                *geom, view, hemi)
            )

    if not opts.get('colorbar', False):
        fig.tight_layout()

    if n_surf == 1:
        fig.subplots_adjust(wspace=-0.1)
    else:
        fig.subplots_adjust(wspace=-0.4, hspace=-0.15)

    return fig


def _fix_facecolors(ax, facecolors, vertices, faces, view, hemi):
    """
    Updates `facecolors` to reflect shading of mesh geometry

    Parameters
    ----------
    ax : plt.Axes3dSubplot
        Axis instance
    facecolors : (F,) array_like
        Original facecolors of plot
    vertices : (V, 3)
        Vertices of surface mesh
    faces : (F, 3)
        Triangles of surface mesh
    view : {'lateral', 'medial'}
        Plotted view of brain

    Returns
    -------
    colors : (F,) array_like
        Updated facecolors with approriate shading
    """

    hemi_view = {'R': {'lateral': 'medial', 'medial': 'lateral'}}
    views = {
        'lateral': plt.cm.colors.LightSource(azdeg=225, altdeg=19.4712),
        'medial': plt.cm.colors.LightSource(azdeg=45, altdeg=19.4712)
    }

    # reverse medial / lateral views if plotting right hemisphere
    view = hemi_view.get(hemi, {}).get(view, view)
    # re-shade colors
    normals = ax._generate_normals(vertices[faces])
    colors = ax._shade_colors(np.asarray(facecolors), normals, views[view])

    return colors



# -*- coding: utf-8 -*-
"""
For testing neuromaps.stats functionality
"""

import numpy as np
import pytest

from neuromaps import stats


@pytest.mark.xfail
def test_compare_images():
    assert False


def test_permtest_metric():
    rs = np.random.default_rng(12345678)
    x, y = rs.random(size=(2, 100))
    r, p = stats.permtest_metric(x, y)
    assert np.allclose([r, p], [0.0345815411043023, 0.7192807192807192])

    r, p = stats.permtest_metric(np.c_[x, x[::-1]], np.c_[y, y])
    assert np.allclose(r, [0.0345815411043023, 0.03338608427980476])
    assert np.allclose(p, [0.7192807192807192, 0.7472527472527473])


@pytest.mark.parametrize('x, y, expected', [
    # basic one-dimensional input
    (range(5), range(5), (1.0, 0.0)),
    # broadcasting occurs regardless of input order
    (np.stack([range(5), range(5, 0, -1)], 1), range(5),
     ([1.0, -1.0], [0.0, 0.0])),
    (range(5), np.stack([range(5), range(5, 0, -1)], 1),
     ([1.0, -1.0], [0.0, 0.0])),
    # correlation between matching columns
    (np.stack([range(5), range(5, 0, -1)], 1),
     np.stack([range(5), range(5, 0, -1)], 1),
     ([1.0, 1.0], [0.0, 0.0]))
])
def test_efficient_pearsonr(x, y, expected):
    assert np.allclose(stats.efficient_pearsonr(x, y), expected)


def test_efficient_pearsonr_errors():
    with pytest.raises(ValueError):
        stats.efficient_pearsonr(range(4), range(5))

    assert all(np.isnan(a) for a in stats.efficient_pearsonr([], []))



# -*- coding: utf-8 -*-
"""
For testing neuromaps.utils functionality
"""

import os
import pytest

from neuromaps import utils


def test_tmpname(tmp_path):
    out = utils.tmpname('.nii.gz', prefix='test', directory=tmp_path)
    assert (isinstance(out, os.PathLike) and out.name.startswith('test')
            and out.name.endswith('.nii.gz'))


@pytest.mark.xfail
def test_run():
    assert False


@pytest.mark.xfail
def test_check_fs_subjid():
    assert False



# -*- coding: utf-8 -*-
"""
For testing neuromaps.images functionality
"""

import nibabel as nib
import numpy as np
import pytest

from neuromaps import images


def test_construct_surf_gii():
    vertices = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 1]])
    tris = np.array([[0, 1, 2]])
    surf = images.construct_surf_gii(vertices, tris)
    assert isinstance(surf, nib.GiftiImage)
    v, t = surf.agg_data()
    assert np.allclose(v, vertices) and np.allclose(t, tris)


@pytest.mark.xfail
def test_construct_shape_gii():
    assert False


@pytest.mark.xfail
def test_fix_coordsys():
    assert False


@pytest.mark.xfail
def test_load_nifti():
    assert False


@pytest.mark.xfail
def test_load_gifti():
    assert False


@pytest.mark.xfail
def test_load_data():
    assert False


@pytest.mark.xfail
def test_obj_to_gifti():
    assert False


@pytest.mark.xfail
def test_fssurf_to_gifti():
    assert False


@pytest.mark.xfail
def test_fsmorph_to_gifti():
    assert False


@pytest.mark.xfail
def test_interp_surface():
    assert False


@pytest.mark.xfail
def test_vertex_areas():
    assert False


@pytest.mark.xfail
def test_average_surfaces():
    assert False


@pytest.mark.xfail
def test_relabel_gifti():
    assert False


@pytest.mark.xfail
def test_annot_to_gifti():
    assert False


@pytest.mark.xfail
def test_dlabel_to_gifti():
    assert False


@pytest.mark.xfail
def test_minc_to_nifti():
    assert False



# -*- coding: utf-8 -*-
"""
Fixtures for all abagen tests
"""

import shutil

import pytest


def pytest_runtest_setup(item):
    markers = set(mark.name for mark in item.iter_markers())
    if 'workbench' in markers and shutil.which('wb_command') is None:
        pytest.skip('Cannot run without Connectome Workbench')



# -*- coding: utf-8 -*-
"""
Functionality for parcellating data
"""

import nibabel as nib
from nilearn.input_data import NiftiLabelsMasker
import numpy as np

from neuromaps.datasets import ALIAS, DENSITIES
from neuromaps.images import construct_shape_gii, load_gifti
from neuromaps.resampling import resample_images
from neuromaps.transforms import _check_hemi
from neuromaps.nulls.spins import vertices_to_parcels, parcels_to_vertices


def _gifti_to_array(gifti):
    """ Converts tuple of `gifti` to numpy array
    """
    return np.hstack([load_gifti(img).agg_data() for img in gifti])


def _array_to_gifti(data):
    """ Converts numpy `array` to tuple of gifti images
    """
    return tuple(construct_shape_gii(arr) for arr in np.split(data, 2))


class Parcellater():
    """
    Class for parcellating arbitrary volumetric / surface data

    Parameters
    ----------
    parcellation : str or os.PathLike or Nifti1Image or GiftiImage or tuple
        Parcellation image or surfaces, where each region is identified by a
        unique integer ID. All regions with an ID of 0 are ignored.
    space : str
        The space in which `parcellation` is defined
    resampling_target : {'data', 'parcellation', None}, optional
        Gives which image gives the final shape/size. For example, if
        `resampling_target` is 'data', the `parcellation` is resampled to the
        space + resolution of the data, if needed. If it is 'parcellation' then
        any data provided to `.fit()` are transformed to the space + resolution
        of `parcellation`. Providing None means no resampling; if spaces +
        resolutions of the `parcellation` and data provided to `.fit()` do not
        match a ValueError is raised. Default: 'data'
    hemi : {'L', 'R'}, optional
        If provided `parcellation` represents only one hemisphere of a surface
        atlas then this specifies which hemisphere. If not specified it is
        assumed that `parcellation` is (L, R) hemisphere. Ignored if `space` is
        'MNI152'. Default: None
    """

    def __init__(self, parcellation, space, resampling_target='data',
                 hemi=None):
        self.parcellation = parcellation
        self.space = ALIAS.get(space, space)
        self.resampling_target = resampling_target
        self.hemi = hemi
        self._volumetric = self.space == 'MNI152'

        if self.resampling_target == 'parcellation':
            self._resampling = 'transform_to_trg'
        else:
            self._resampling = 'transform_to_src'

        if not self._volumetric:
            self.parcellation, self.hemi = zip(
                *_check_hemi(self.parcellation, self.hemi)
            )

        if self.resampling_target not in ('parcellation', 'data', None):
            raise ValueError('Invalid value for `resampling_target`: '
                             f'{resampling_target}')

        if self.space not in DENSITIES:
            raise ValueError(f'Invalid value for `space`: {space}')

    def fit(self):
        """ Prepare parcellation for data extraction
        """

        if not self._volumetric:
            self.parcellation = tuple(
                load_gifti(img) for img in self.parcellation
            )
        self._fit = True
        return self

    def transform(self, data, space, hemi=None):
        """
        Applies parcellation to `data` in `space`

        Parameters
        ----------
        data : str or os.PathLike or Nifti1Image or GiftiImage or tuple
            Data to parcellate
        space : str
            The space in which `data` is defined
        hemi : {'L', 'R'}, optional
            If provided `data` represents only one hemisphere of a surface
            dataset then this specifies which hemisphere. If not specified it
            is assumed that `data` is (L, R) hemisphere. Ignored if `space` is
            'MNI152'. Default: None

        Returns
        -------
        parcellated : np.ndarray
            Parcellated `data`
        """

        self._check_fitted()

        space = ALIAS.get(space, space)
        if (self.resampling_target == 'data' and space == 'MNI152'
                and not self._volumetric):
            raise ValueError('Cannot use resampling_target="data" when '
                             'provided parcellation is in surface space and '
                             'provided data are in MNI1512 space.')
        elif (self.resampling_target == 'parcellation' and self._volumetric
                and space != 'MNI152'):
            raise ValueError('Cannot use resampling_target="parcellation" '
                             'when provided parcellation is in MNI152 space '
                             'and provided are in surface space.')

        if hemi is not None and hemi not in self.hemi:
            raise ValueError('Cannot parcellate data from {hemi} hemisphere '
                             'when parcellation was provided for incompatible '
                             'hemisphere: {self.hemi}')

        if isinstance(data, np.ndarray):
            data = _array_to_gifti(data)
        data, parc = resample_images(data, self.parcellation,
                                     space, self.space, hemi=hemi,
                                     resampling=self._resampling,
                                     method='nearest')

        if ((self.resampling_target == 'data'
             and space.lower() == 'mni152')
                or (self.resampling_target == 'parcellation'
                    and self._volumetric)):
            data = nib.concat_images([nib.squeeze_image(data)])
            parcellated = NiftiLabelsMasker(
                parc, resampling_target=self.resampling_target
            ).fit_transform(data)
        else:
            if not self._volumetric:
                for n, _ in enumerate(parc):
                    parc[n].labeltable.labels = \
                        self.parcellation[n].labeltable.labels
            data = _gifti_to_array(data)
            parcellated = vertices_to_parcels(data, parc)

        return parcellated

    def inverse_transform(self, data):
        """
        Project `data` to space + density of parcellation

        Parameters
        ----------
        data : array_like
            Parcellated data to be projected to the space of parcellation

        Returns
        -------
        data : Nifti1Image or tuple-of-nib.GiftiImage
            Provided `data` in space + resolution of parcellation
        """

        if not self._volumetric:
            verts = parcels_to_vertices(data, self.parcellation, self.drop)
            img = _array_to_gifti(verts)
        else:
            data = np.atleast_2d(data)
            img = NiftiLabelsMasker(self.parcellation).fit() \
                                                      .inverse_transform(data)
        return img

    def fit_transform(self, data, space, hemi=None):
        """ Prepare and perform parcellation of `data`
        """
        return self.fit().transform(data, space, hemi)

    def _check_fitted(self):
        if not hasattr(self, '_fit'):
            raise ValueError(f'It seems that {self.__class__.__name__} has '
                             'not been fit. You must call `.fit()` before '
                             'calling `.transform()`')



# -*- coding: utf-8 -*-
"""
For testing neuromaps.nulls.spins functionality
"""

import numpy as np
import pytest

from neuromaps.nulls import spins


def test_load_spins():
    out = np.random.randint(1000, size=(100, 100), dtype='int32')
    assert out is spins.load_spins(out)
    assert np.allclose(out[:, :10], spins.load_spins(out, n_perm=10))


@pytest.mark.xfail
def test_get_parcel_centroids():
    assert False


@pytest.mark.xfail
def test__gen_rotation():
    assert False


@pytest.mark.xfail
def test_gen_spinsamples():
    assert False


@pytest.mark.xfail
def test_spin_parcels():
    assert False


@pytest.mark.xfail
def test_parcels_to_vertices():
    assert False


@pytest.mark.xfail
def test_vertices_to_parcels():
    assert False


@pytest.mark.xfail
def test_spin_data():
    assert False



# -*- coding: utf-8 -*-
"""
Utilites for loading / creating datasets
"""

import json
import os
from pkg_resources import resource_filename

import requests

RESTRICTED = ["grh4d"]


def _osfify_urls(data, return_restricted=True):
    """
    Formats `data` object with OSF API URL

    Parameters
    ----------
    data : object
        If dict with a `url` key, will format OSF_API with relevant values
    return_restricted : bool, optional
        Whether to return restricted annotations. These will only be accesible
        with a valid OSF token. Default: True

    Returns
    -------
    data : object
        Input data with all `url` dict keys formatted
    """

    OSF_API = "https://files.osf.io/v1/resources/{}/providers/osfstorage/{}"

    if isinstance(data, str) or data is None:
        return data
    elif 'url' in data:
        # if url is None then we this is a malformed entry and we should ignore
        if data['url'] is None:
            return
        # if the url isn't a string assume we're supposed to format it
        elif not isinstance(data['url'], str):
            if data['url'][0] in RESTRICTED and not return_restricted:
                return
            data['url'] = OSF_API.format(*data['url'])

    try:
        for key, value in data.items():
            data[key] = _osfify_urls(value, return_restricted)
    except AttributeError:
        for n, value in enumerate(data):
            data[n] = _osfify_urls(value, return_restricted)
        # drop the invalid entries
        data = [d for d in data if d is not None]

    return data


def get_dataset_info(name, return_restricted=True):
    """
    Returns information for requested dataset `name`

    Parameters
    ----------
    name : str
        Name of dataset
    return_restricted : bool, optional
        Whether to return restricted annotations. These will only be accesible
        with a valid OSF token. Default: True

    Returns
    -------
    dataset : dict or list-of-dict
        Information on requested data
    """

    fn = resource_filename('neuromaps',
                           os.path.join('datasets', 'data', 'osf.json'))
    with open(fn) as src:
        osf_resources = _osfify_urls(json.load(src), return_restricted)

    try:
        resource = osf_resources[name]
    except KeyError:
        raise KeyError("Provided dataset '{}' is not valid. Must be one of: {}"
                       .format(name, sorted(osf_resources.keys())))

    return resource


def get_data_dir(data_dir=None):
    """
    Gets path to neuromaps data directory

    Parameters
    ----------
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NEUROMAPS_DATA'; if that is not set, will
        use `~/neuromaps-data` instead. Default: None

    Returns
    -------
    data_dir : str
        Path to use as data directory
    """

    if data_dir is None:
        data_dir = os.environ.get('NEUROMAPS_DATA',
                                  os.path.join('~', 'neuromaps-data'))
    data_dir = os.path.expanduser(data_dir)
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    return data_dir


def _get_token(token=None):
    """
    Returns `token` if provided or set as environmental variable

    Parameters
    ----------
    token : str, optional
        OSF personal access token for accessing restricted annotations. Will
        also check the environmental variable 'NEUROMAPS_OSF_TOKEN' if not
        provided; if that is not set no token will be provided and restricted
        annotations will be inaccessible. Default: None

    Returns
    -------
    token : str
        OSF token
    """

    if token is None:
        token = os.environ.get('NEUROMAPS_OSF_TOKEN', None)

    return token


def _get_session(token=None):
    """
    Returns requests.Session with `token` auth in header if supplied

    Parameters
    ----------
    token : str, optional
        OSF personal access token for accessing restricted annotations. Will
        also check the environmental variable 'NEUROMAPS_OSF_TOKEN' if not
        provided; if that is not set no token will be provided and restricted
        annotations will be inaccessible. Default: None

    Returns
    -------
    session : requests.Session
        Session instance with authentication in header
    """

    session = requests.Session()
    token = _get_token(token)
    if token is not None:
        session.headers['Authorization'] = 'Bearer {}'.format(token)

    return session



# -*- coding: utf-8 -*-
"""
For testing neuromaps.nulls.nulls functionality
"""

import pytest


@pytest.mark.xfail
def test_alexander_bloch():
    assert False


@pytest.mark.xfail
def test_vasa():
    assert False


@pytest.mark.xfail
def test_hungarian():
    assert False


@pytest.mark.xfail
def test_baum():
    assert False


@pytest.mark.xfail
def test_cornblath():
    assert False


@pytest.mark.xfail
def test__get_distmat():
    assert False


@pytest.mark.xfail
def test__make_surrogates():
    assert False


@pytest.mark.xfail
def test_burt2018():
    assert False


@pytest.mark.xfail
def test_burt2020():
    assert False


@pytest.mark.xfail
def test_moran():
    assert False



# -*- coding: utf-8 -*-
"""
For testing neuromaps.datasets.atlases functionality
"""

import pytest

from neuromaps.datasets import atlases


@pytest.mark.parametrize('atlas, expected', [
    ('fslr', 'fsLR'), ('fsLR', 'fsLR'), ('fsavg', 'fsaverage'),
    ('fsaverage', 'fsaverage'), ('CIVET', 'civet'), ('civet', 'civet'),
    ('mni152', 'MNI152'), ('mni', 'MNI152'), ('MNI152', 'MNI152')
])
def test__sanitize_atlas(atlas, expected):
    assert atlases._sanitize_atlas(atlas) == expected


def test__sanitize_atlas_errors():
    with pytest.raises(ValueError):
        atlases._sanitize_atlas('invalid')


@pytest.mark.xfail
def test__bunch_outputs():
    assert False


@pytest.mark.xfail
def test__fetch_atlas():
    assert False


@pytest.mark.xfail
def test_fetch_civet():
    assert False


@pytest.mark.xfail
def test_fetch_fsaverage():
    assert False


@pytest.mark.xfail
def test_fetch_mni152():
    assert False


@pytest.mark.xfail
def test_fetch_regfusion():
    assert False


@pytest.mark.xfail
def test_fetch_atlas():
    assert False


@pytest.mark.xfail
def test_fetch_all_atlases():
    assert False


@pytest.mark.xfail
def test_get_atlas_dir():
    assert False



# -*- coding: utf-8 -*-
"""
Functions for working with data/osf.json file
"""

import os
from pkg_resources import resource_filename
import json

from nilearn.datasets.utils import _md5_sum_file

from neuromaps.datasets.utils import _get_session

# uniquely identify each item ('hemi' can be None)
FNAME_KEYS = ['source', 'desc', 'space', 'den', 'res', 'hemi']
# auto-generated (checksum can be None if file doest not exist)
AUTO_KEYS = ['format', 'fname', 'rel_path', 'checksum']
# required keys but values are all optional
COND_KEYS = ['title', 'tags', 'redir', 'url']
# minimal keys for each item
MINIMAL_KEYS = FNAME_KEYS + AUTO_KEYS + COND_KEYS
# keys for redirection
REDIR_KEYS = ['space', 'den']
# keys for more metadata (unique for each source)
INFO_KEYS = ['source', 'refs', 'comments', 'demographics']

# distribution JSON
OSFJSON = resource_filename(
    'neuromaps', os.path.join('datasets', 'data', 'osf.json')
)


def parse_filename(fname, return_ext=True, verbose=False):
    """
    Parses `fname` (in BIDS-inspired format) and returns dictionary

    Parameters
    ----------
    fname : str os os.PathLike
        Filename to parse
    return_ext : bool, optional
        Whether to return extension of `fname` in addition to key-value dict.
        Default: False
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    info : dict
        Key-value pairs extracted from `fname`
    ext : str
        Extension of `fname`, only returned if `return_ext=True`
    """

    try:
        base, *ext = fname.split('.')
        fname_dict = dict([
            pair.split('-') for pair in base.split('_') if pair != 'feature'
        ])
    except ValueError:
        print('Wrong filename format!')
        return

    if verbose:
        print(fname_dict)

    if return_ext:
        return fname_dict, '.'.join(ext)

    return fname_dict


def parse_fname_list(fname, verbose=False):
    """
    Reads in list of BIDS-inspired filenames from `fname` and parses keys

    Parameters
    ----------
    fname : str os os.PathLike
    verbose : bool, optional
        Whether to print status messages. Default: False

    Returns
    -------
    data : list-of-dict
        Information about filenames in `fname`
    """

    with open(fname, 'r', encoding='utf-8') as src:
        fname_list = [name.strip() for name in src.readlines()]
    data = [
        parse_filename(name, return_ext=False, verbose=verbose)
        for name in fname_list
    ]
    if verbose:
        print(fname_list)

    return data


def parse_json(fname, root='annotations'):
    """
    Loads JSON from `fname` and returns value of `root` key(s)

    Parameters
    ----------
    fname : str or os.PathLike
        Filepath to JSON file
    root : str or list-of-str, optional
        Root key(s) to query JSON file. Default: 'annotations'

    Returns
    -------
    data : dict
        Data from `fname` JSON file
    """

    if isinstance(root, str):
        root = [root]

    with open(fname, 'r', encoding='utf-8') as src:
        data = json.load(src)

    for key in root:
        data = data[key]

    return data


def write_json(data, fname, root='annotations', indent=4):
    """
    Saves `data` to `fname` JSON

    Parameters
    ----------
    data : JSON-compatible format
        Data to save to `fname`
    fname : str or os.PathLike
        Path to filename where `data` should be saved as JSON
    root : str, optional
        Key to save `data` in `fname`. Default: 'annotations'
    indent : int, optional
        Indentation of JSON file. Default: 4

    Returns
    -------
    fname : str
        Path to saved file
    """

    if not isinstance(root, str):
        raise ValueError(f'Provided `root` must be a str. Received: {root}')

    # if `fname` already exists we want to update it, not overwrite it!
    output = {root: []}
    if os.path.isfile(fname) and os.stat(fname).st_size != 0:
        output = parse_json(fname, [])
    if output.get(root) is None:
        output[root] = []

    # update entries in output json if they already exist; otherwise, overwrite
    keys = ('fname', 'rel_path', 'checksum')
    missing = []
    for n, entry in enumerate(output[root]):
        comp = {k: entry[k] for k in keys}
        match = False
        for item in data:
            red = {k: item[k] for k in keys}
            if comp == red:
                output[root][n] = item
                match = True
        if not match:
            missing.append(item)

    output[root].extend(missing)

    # save to disk
    with open(fname, 'w', encoding='utf-8') as dest:
        json.dump(output, dest, indent=indent)

    return fname


def complete_json(input_data, ref_keys='minimal', input_root=None,
                  output_fname=None, output_root=None):
    """
    Parameters
    ----------
    input_data : str or os.PathLike or list-of-dict
        Filepath to JSON with data or list of dictionaries with information
        about annotations
    ref_keys : {'minimal', 'info'}, optional
        Which reference keys to check in `input_data`. Default: 'minimal'
    input_root : str, optional
        If `input_data` is a filename the key in the file containing data about
        annotations. If not specified will be based on provided `ref_keys`.
        Default: None
    output_fname : str or os.PathLike, optional
        Filepath where complete JSON should be saved. If not specified the
        data are not saved to disk. Default: None
    output_root : str, optional
        If `output_fname` is not None, the key in the saved JSON where
        completed information should be stored. If not specified will be based
        on `input_root`. Default: None

    Returns
    -------
    output : list-of-dict
        Information about annotations from `input_data`
    """
    valid_keys = ['minimal', 'info']
    if ref_keys not in valid_keys:
        raise ValueError(f'Invalid ref_keys: {ref_keys}. Must be one of '
                         f'{valid_keys}')

    # this is to add missing fields to existing data
    # could accept data dict list or filename as input
    # set minimal vs info
    if ref_keys == 'minimal':
        ref_keys = MINIMAL_KEYS
        if input_root is None:
            input_root = 'annotations'
    elif ref_keys == 'info':
        ref_keys = INFO_KEYS
        if input_root is None:
            input_root = 'info'

    # check input
    if not isinstance(input_data, list):
        input_data = parse_json(input_data, root=input_root)

    # make output
    output = []
    for item in input_data:
        output.append({
            key: (item[key] if key in item else None)
            for key in ref_keys
        })

    # write output
    if output_fname is not None:
        if output_root is None:
            output_root = input_root
        write_json(output, output_fname, root=output_root)

    return output


def check_missing_keys(fname, root='annotations'):
    """
    Checks whether data in `fname` JSON are missing required keys

    Required keys are specified in ``neuromaps.datasets._osf.MINIMAL_KEYS``

    Parameters
    ----------
    fname : str or os.PathLike
        Filepath to JSON file to check
    root : str or list-of-str, optional
        Root key(s) to query JSON file. Default: 'annotations'

    Returns
    -------
    info : list of list-of-str
        Missing keys for each entry in `fname`
    """

    try:
        data = parse_json(fname, root=root)
    except TypeError:
        if isinstance(fname, dict):
            data = [fname]

    is_missing_keys, info = False, []
    for item in data:
        missing = sorted(set(MINIMAL_KEYS) - set(item))
        if len(missing) > 0:
            is_missing_keys = True
        info.append(missing)

    if is_missing_keys:
        raise KeyError('Data in provided `fname` are missing some keys. '
                       'Please use `neuromaps.datasets._osf.complete_json`'
                       ' to fill missing keys')

    return info


def generate_auto_keys(item):
    """
    Adds automatically-generated keys to `item`

    Generated keys include: ['format', 'fname', 'rel_path', 'checksum']

    Parameters
    ----------
    item : dict
        Information about annotation

    Returns
    -------
    item : dict
        Updated information about annotation
    """

    item = item.copy()

    pref = 'source-{source}_desc-{desc}_space-{space}'
    surffmt = pref + '_den-{den}_hemi-{hemi}_feature.func.gii'
    volfmt = pref + '_res-{res}_feature.nii.gz'

    # check format by checking 'hemi'
    is_surface = ('den' in item or 'hemi' in item
                  or item.get('format') == 'surface')
    is_volume = 'res' in item or item.get('format') == 'volume'

    if is_surface:  # this is surface file
        item['format'] = 'surface'
        item['fname'] = surffmt.format(**item)
    elif is_volume:  # this is volume file
        item['format'] = 'volume'
        item['fname'] = volfmt.format(**item)
    else:
        raise ValueError('Missing keys to determine surface/volumetric format '
                         'of data')

    item['rel_path'] = os.path.join(*[
        item[key] for key in ['source', 'desc', 'space']
    ])

    # check file existence
    filepath = os.path.join(item['rel_path'], item['fname'])
    if item['fname'] is not None and os.path.isfile(filepath):
        if item.get('checksum') is None:
            item['checksum'] = _md5_sum_file(filepath)

    return item


def clean_minimal_keys(item):
    """
    Removes incompatible keys from `item` based on `item['format']`

    Parameters
    ----------
    item : dict
        Information about annotation

    Returns
    -------
    item : dict
        Updated information about annotation
    """

    keys = {'surface': ['res'], 'volume': ['den', 'hemi']}
    fmt = item.get('format')
    if fmt is None:
        print('Invalid value for format key; setting to "null"')
        item['format'] = None
        return

    for key in keys.get(fmt, []):
        item.pop(key, None)

    return item


def get_url(fname, project, token=None):
    """
    Gets OSF API URL path for `fname` in `project`

    Parameters
    ----------
    fname : str
        Filepath as it exists on OSF
    project : str
        Project ID on OSF
    token : str, optional
        OSF personal access token for accessing restricted annotations. Will
        also check the environmental variable 'NEUROMAPS_OSF_TOKEN' if not
        provided; if that is not set no token will be provided and restricted
        annotations will be inaccessible. Default: None

    Returns
    -------
    path : str
        Path to `fname` on OSF project `project`
    """

    url = f'https://files.osf.io/v1/resources/{project}/providers/osfstorage/'
    session = _get_session(token=token)
    path = ''
    for pathpart in fname.strip('/').split('/'):
        out = session.get(url + path)
        out.raise_for_status()
        for item in out.json()['data']:
            if item['attributes']['name'] == pathpart:
                break
        path = item['attributes']['path'][1:]

    return path


def generate_release_json(fname, output=OSFJSON, root='annotations',
                          project=None, token=None):
    """
    Generates distribution-ready JSON file for fetching annotation data

    Parameters
    ----------
    fname : str or os.PathLike
        Path to filename where manually-edited JSON information is stored
    output : str or os.PathLike
        Path to filename where output JSON should be saved
    root : str, optional
        Key in `fname` where relevant data are stored. Default: 'annotations'
    project : str, optional
        Project ID on OSF where data files are stored. If not specified then
        the URL for the generated data will not be set. Default: None
    token : str, optional
        OSF personal access token for accessing restricted annotations. Will
        also check the environmental variable 'NEUROMAPS_OSF_TOKEN' if not
        provided; if that is not set no token will be provided and restricted
        annotations will be inaccessible. Default: None

    Returns
    -------
    output : str
        Path to filename where output JSON was saved
    """

    info = []
    for item in parse_json(fname, root=root):
        item = clean_minimal_keys(generate_auto_keys(item))
        # fetch URL for file if needed (and project is specified)
        if (item.get('fname') is not None and item.get('url') is None
                and project is not None):
            fn = os.path.join(root, item['rel_path'], item['fname'])
            item['url'] = [project, get_url(fn, project=project, token=token)]
        info.append({key: item[key] for key in MINIMAL_KEYS if key in item})
    fname = write_json(info, output, root='annotations')
    return fname



# -*- coding: utf-8 -*-
"""
For testing neuromaps.nulls.burt functionality
"""

import numpy as np
import pytest

from neuromaps.nulls import burt


def test__make_weight_matrix():
    x0 = np.random.rand(100, 100)
    out = burt._make_weight_matrix(x0, 0.5)
    assert out.shape == x0.shape
    assert np.allclose(np.diag(out), 0)


@pytest.mark.xfail
def test_estimate_rho_d0():
    assert False


@pytest.mark.xfail
def test_make_surrogate():
    assert False


@pytest.mark.xfail
def test_batch_surrogates():
    assert False



# -*- coding: utf-8 -*-
"""
Functions for comparing data
"""

import nibabel as nib
import numpy as np

from neuromaps import transforms
from neuromaps.datasets import ALIAS, DENSITIES
from neuromaps.images import load_gifti, load_nifti


_resampling_docs = dict(
    resample_in="""\
src, trg : str or os.PathLike or niimg_like or nib.GiftiImage or tuple
    Input data to be resampled
src_space, trg_space : str
    Template space of input data
method : {'nearest', 'linear'}, optional
    Method for resampling. Specify 'nearest' if `data` are label images.
    Default: 'linear'\
""",
    hemi="""\
hemi : {'L', 'R'}, optional
    If `src` and `trg` are not tuples this specifies the hemisphere the data
    represent. Default: None\
""",
    resample_out="""\
src, trg : tuple-of-nib.GiftiImage
    Resampled images\
"""
)


def downsample_only(src, trg, src_space, trg_space, method='linear',
                    hemi=None):
    src_den, trg_den = transforms._estimate_density((src, trg), hemi)
    src_num, trg_num = int(src_den[:-1]), int(trg_den[:-1])
    src_space, trg_space = src_space.lower(), trg_space.lower()

    if src_num >= trg_num:  # resample to `trg`
        func = getattr(transforms, f'{src_space}_to_{trg_space}')
        src = func(src, trg_den, hemi=hemi, method=method)
    elif src_num < trg_num:  # resample to `src`
        func = getattr(transforms, f'{trg_space}_to_{src_space}')
        trg = func(trg, src_den, hemi=hemi, method=method)

    return src, trg


downsample_only.__doc__ = """\
Resamples `src` and `trg` to match such that neither is upsampled

If density of `src` is greater than `trg` then `src` is resampled to
`trg`; otherwise, `trg` is resampled to `src`

Parameters
----------
{resample_in}
{hemi}

Returns
-------
{resample_out}
""".format(**_resampling_docs)


def transform_to_src(src, trg, src_space, trg_space, method='linear',
                     hemi=None):
    src_den, trg_den = transforms._estimate_density((src, trg), hemi)

    func = getattr(transforms, f'{trg_space.lower()}_to_{src_space.lower()}')
    trg = func(trg, src_den, hemi=hemi, method=method)

    return src, trg


transform_to_src.__doc__ = """\
Resamples `trg` to match space and density of `src`

Parameters
----------
{resample_in}
{hemi}

Returns
-------
{resample_out}
""".format(**_resampling_docs)


def transform_to_trg(src, trg, src_space, trg_space, hemi=None,
                     method='linear'):
    src_den, trg_den = transforms._estimate_density((src, trg), hemi)

    func = getattr(transforms, f'{src_space.lower()}_to_{trg_space.lower()}')
    src = func(src, trg_den, hemi=hemi, method=method)

    return src, trg


transform_to_trg.__doc__ = """\
Resamples `trg` to match space and density of `src`

Parameters
----------
{resample_in}

Returns
-------
{resample_out}
""".format(**_resampling_docs)


def transform_to_alt(src, trg, src_space, trg_space, method='linear',
                     hemi=None, alt_space='fsaverage', alt_density='41k'):
    src, _ = transform_to_trg(src, alt_density, src_space, alt_space,
                              hemi=hemi, method=method)
    trg, _ = transform_to_trg(trg, alt_density, trg_space, alt_space,
                              hemi=hemi, method=method)

    return src, trg


transform_to_alt.__doc__ = """\
Resamples `src` and `trg` to `alt_space` and `alt_density`

Parameters
----------
{resample_in}
{hemi}
alt_space : {{'fsaverage', 'fsLR', 'civet'}}, optional
    Alternative space to which `src` and `trg` should be transformed. Default:
    'fsaverage'
alt_density : str, optional
    Resolution to which `src` and `trg` should be resampled. Must be valid
    with `alt_space`. Default: '41k'

Returns
-------
{resample_out}
""".format(**_resampling_docs)


def mni_transform(src, trg, src_space, trg_space, method='linear', hemi=None):
    if src_space != 'MNI152':
        raise ValueError('Cannot perform MNI transformation when src_space is '
                         f'not "MNI152." Received: {src_space}.')
    trg_den = trg
    if trg_space != 'MNI152':
        trg_den, = transforms._estimate_density((trg_den,), hemi)
    func = getattr(transforms, f'mni152_to_{trg_space.lower()}')
    src = func(src, trg_den, method=method)

    return src, trg


mni_transform.__doc__ = """\
Resamples `src` in MNI152 to `trg` space

Parameters
----------
{resample_in}
hemi : {{'L', 'R'}}, optional
    If `trg_space` is not "MNI152' and `trg` is not a tuple this specifies the
    hemisphere the data represent. Default: None

Returns
-------
{resample_out}
""".format(**_resampling_docs)


def _check_altspec(spec):
    """
    Confirms that specified alternative `spec` is valid (space, density) format

    Parameters
    ----------
    spec : (2,) tuple-of-str
        Where entries are (space, density) of desired target space

    Returns
    -------
    spec : (2,) tuple-of-str
        Unmodified input `spec`

    Raises
    ------
    ValueError
        If `spec` is not valid format
    """

    invalid_spec = spec is None or len(spec) != 2
    if not invalid_spec:
        space, den = spec
        space = ALIAS.get(space, space)
        valid = DENSITIES.get(space)
        invalid_spec = valid is None or den not in valid
    if invalid_spec:
        raise ValueError('Must provide valid alternative specification of '
                         f'format (space, density). Received: {spec}')

    return (space, den)


def resample_images(src, trg, src_space, trg_space, method='linear',
                    hemi=None, resampling='downsample_only', alt_spec=None):

    resamplings = ('downsample_only', 'transform_to_src', 'transform_to_trg',
                   'transform_to_alt')
    if resampling not in resamplings:
        raise ValueError(f'Invalid method: {resampling}')

    src_space = ALIAS.get(src_space, src_space)
    trg_space = ALIAS.get(trg_space, trg_space)

    # all this input handling just to deal with volumetric images :face_palm:
    opts, err = {}, None
    if resampling == 'transform_to_alt':
        opts['alt_space'], opts['alt_density'] = _check_altspec(alt_spec)
        if (opts['alt_space'] == 'MNI152'
                and (src_space != 'MNI152' or trg_space != 'MNI152')):
            raise ValueError('Cannot transform to "MNI152" system if either '
                             '`src` or `trg` are not in "MNI152" system.')
    elif (resampling == 'transform_to_src' and src_space == 'MNI152'
            and trg_space != 'MNI152'):
        err = ('Specified `src_space` cannot be "MNI152" when `resampling` is '
               '"transform_to_src"')
    elif (resampling == 'transform_to_trg' and src_space != 'MNI152'
            and trg_space == 'MNI152'):
        err = ('Specified `trg_space` cannot be "MNI152" when `resampling` is '
               '"transform_to_trg"')
    elif (resampling == 'transform_to_alt' and opts['alt_space'] == 'MNI152'
            and (src_space != 'MNI152' or trg_space != 'MNI152')):
        err = ('Specified `alt_space` cannot be "MNI152" when `resampling` is '
               '"transform_to_alt"')
    if err is not None:
        raise ValueError(err)

    # handling volumetric data is annoying...
    if ((src_space == "MNI152" or trg_space == "MNI152")
            and resampling == 'transform_to_alt'):
        func = mni_transform if src_space == 'MNI152' else transform_to_trg
        src = func(src, opts['alt_density'], src_space, opts['alt_space'],
                   method=method, hemi=hemi)[0]
        func = mni_transform if trg_space == 'MNI152' else transform_to_trg
        trg = func(trg, opts['alt_density'], trg_space, opts['alt_space'],
                   method=method, hemi=hemi)[0]
    elif src_space == 'MNI152' and trg_space != 'MNI152':
        src, trg = mni_transform(src, trg, src_space, trg_space,
                                 method=method, hemi=hemi)
    elif trg_space == 'MNI152' and src_space != 'MNI152':
        trg, src = mni_transform(trg, src, trg_space, src_space,
                                 method=method, hemi=hemi)
    elif src_space == 'MNI152' and src_space == 'MNI152':
        src, trg = load_nifti(src), load_nifti(trg)
        srcres = np.prod(nib.affines.voxel_sizes(src.affine))
        trgres = np.prod(nib.affines.voxel_sizes(trg.affine))
        if ((resampling == 'downsample_only' and srcres > trgres)
                or resampling == 'transform_to_src'):
            trg, src = mni_transform(trg, src, trg_space, src_space,
                                     method=method)
        elif ((resampling == 'downsample_only' and srcres <= trgres)
                or resampling == 'transform_to_trg'):
            src, trg = mni_transform(src, trg, src_space, trg_space,
                                     method=method)
    else:
        func = globals()[resampling]
        src, trg = func(src, trg, src_space, trg_space, hemi=hemi,
                        method=method, **opts)
        src = tuple(load_gifti(s) for s in src)
        trg = tuple(load_gifti(t) for t in trg)

    return src, trg


resample_images.__doc__ = """\
Resamples images `src` and `trg` to same space/density with `resampling` method

Parameters
----------
{resample_in}
{hemi}
resampling : str, optional
    Name of resampling function to resample `src` and `trg`. Must be one of:
    'downsample_only', 'transform_to_src', 'transform_to_trg',
    'transform_to_alt'. See Notes for more info. Default: 'downsample_only'
alt_spec : (2,) tuple-of-str
    Where entries are (space, density) of desired target space. Only used if
    `resampling='transform_to_alt'`. Default: None

Returns
-------
{resample_out}

Notes
-----
The four available `resampling` strategies will control how `src` and/or `trg`
are resampled prior to correlation. Options include:

    1. `resampling='downsample_only'`

    Data from `src` and `trg` are resampled to the lower resolution of the two
    input datasets

    2. `resampling='transform_to_src'`

    Data from `trg` are always resampled to match `src` space and resolution

    3. `resampling='transform_to_trg'`

    Data from `src` are always resampled to match `trg` space and resolution

    4. `resampling='transform_to_alt'`

    Data from `trg` and `src` are resampled to the space and resolution
    specified by `alt_spec` (space, density)

""".format(**_resampling_docs)



# -*- coding: utf-8 -*-
"""
Functions for statistical analyses
"""

from functools import partial

import numpy as np
from scipy import special, stats as sstats
from scipy.stats.stats import _chk2_asarray
from sklearn.utils.validation import check_random_state

from neuromaps.images import load_data


def compare_images(src, trg, metric='pearsonr', ignore_zero=True, nulls=None,
                   nan_policy='omit'):
    """
    Compares images `src` and `trg`

    If `src` and `trg` represent data from multiple hemispheres the data are
    concatenated across hemispheres prior to comparison

    Parameters
    ----------
    src, trg : str or os.PathLike or nib.GiftiImage or niimg_like or tuple
        Images to be compared
    metric : {'pearsonr', 'spearmanr', callable}, optional
        Type of similarity metric to use to compare `src` and `trg` images. If
        a callable function is provided it must accept two inputs and return a
        single value (the similarity metric). Default: 'pearsonr'
    ignore_zero : bool, optional
        Whether to perform comparisons ignoring all zero values in `src` and
        `trg` data. Default: True
    nulls : array_like, optional
        Null data for `src` to use in generating a non-parametric p-value.
        If not specified a parameteric p-value is generated. Default: None
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle when input contains nan. 'propagate' returns nan,
        'raise' throws an error, 'omit' performs the calculations ignoring nan
        values. Default: 'omit'

    Returns
    -------
    similarity : float
         Comparison metric between `src` and `trg`
    pvalue : float
        The p-value of `similarity`, if `nulls` is not None
    """

    methods = ('pearsonr', 'spearmanr')
    if metric not in methods:
        if not callable(metric):
            raise ValueError(f'Invalid `metric`: {metric}')
        else:
            if not isinstance(metric([1, 1], [1, 1]), float):
                raise ValueError('Provided callable `metric` must accept two '
                                 'inputs and return single value.')

    srcdata, trgdata = load_data(src), load_data(trg)

    mask = np.zeros(len(srcdata), dtype=bool)
    if ignore_zero:
        mask = np.logical_or(np.isclose(srcdata, 0), np.isclose(trgdata, 0))

    # drop NaNs
    nanmask = np.logical_or(np.isnan(srcdata), np.isnan(trgdata))
    if np.any(nanmask) and nan_policy == 'raise':
        raise ValueError('Inputs contain nan')
    mask = np.logical_and(np.logical_not(mask), np.logical_not(nanmask))
    srcdata, trgdata = srcdata[mask], trgdata[mask]

    if metric in methods:
        if metric == 'spearmanr':
            srcdata = sstats.rankdata(srcdata)
            trgdata = sstats.rankdata(trgdata)
        metric = partial(efficient_pearsonr, return_pval=False)

    if nulls is not None:
        n_perm = nulls.shape[-1]
        nulls = nulls[mask]
        return permtest_metric(srcdata, trgdata, metric, n_perm=n_perm,
                               nulls=nulls, nan_policy=nan_policy)

    return metric(srcdata, trgdata)


def permtest_metric(a, b, metric='pearsonr', n_perm=1000, seed=0, nulls=None,
                    nan_policy='propagate'):
    """
    Generates non-parameteric p-value of `a` and `b` for `metric`

    Calculates two-tailed p-value for hypothesis of whether samples `a` and `b`
    are related using permutation tests

    Parameters
    ----------
    a, b : (N,) array_like
        Sample observations. These arrays must have the same length
    metric : {'pearsonr', 'spearmanr', callable}, optional
        Type of similarity metric to use to compare `a` and `b`. If a callable
        function is provided it must accept two inputs and return a single
        value (the similarity metric). Default: 'pearsonr'
    n_perm : int, optional
        Number of permutations to assess. Unless `a` and `b` are very small
        this will approximate a randomization test via Monte Carlo simulations.
        Default: 1000
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Set to None for pseudo-randomness.
        Default: 0
    nulls : (N, P) array_like, optional
        Null array used in place of shuffled `a` array to compute null
        distribution of correlations. Array must have the same length as `a`
        and `b`. Providing this will override the value supplied to `n_perm`.
        When not specified a standard permutation is used to shuffle `a`.
        Default: None
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle when inputs contain nan. 'propagate' returns nan,
        'raise' throws an error, 'omit' performs the calculations ignoring nan
        values. Default: 'propagate'

    Returns
    -------
    similarity : float
        Similarity metric
    pvalue : float
        Non-parametric p-value

    Notes
    -----
    The lowest p-value that can be returned by this function is equal to 1 /
    (`n_perm` + 1).
    """

    def nan_wrap(a, b, nan_policy='propagate'):
        nanmask = np.logical_or(np.isnan(a), np.isnan(b))
        if nan_policy == 'raise':
            if np.any(nanmask):
                raise ValueError('Input contains nan')
        elif nan_policy == 'omit':
            a, b = a[~nanmask], b[~nanmask]
        return metric(a, b)

    a, b, _ = _chk2_asarray(a, b, 0)
    rs = check_random_state(seed)

    if len(a) != len(b):
        raise ValueError('Provided arrays do not have same length')

    if a.size == 0 or b.size == 0:
        return np.nan, np.nan

    methods = ('pearsonr', 'spearmanr')
    if metric in methods:
        if metric == 'spearmanr':
            a, b = sstats.rankdata(a), sstats.rankdata(b)
        compfunc = partial(efficient_pearsonr, return_pval=False)
    else:
        compfunc = nan_wrap

    if nulls is not None:
        n_perm = nulls.shape[-1]

    # divide by one forces coercion to float if ndim = 0
    true_sim = compfunc(a, b, nan_policy=nan_policy) / 1
    abs_true = np.abs(true_sim)

    permutations = np.ones(true_sim.shape)
    for perm in range(n_perm):
        # permute `a` and determine whether correlations exceed original
        ap = a[rs.permutation(len(a))] if nulls is None else nulls[:, perm]
        permutations += np.abs(
            compfunc(ap, b, nan_policy=nan_policy)
        ) >= abs_true

    pvals = permutations / (n_perm + 1)  # + 1 in denom accounts for true_sim

    return true_sim, pvals


def efficient_pearsonr(a, b, ddof=1, nan_policy='propagate', return_pval=True):
    """
    Computes correlation of matching columns in `a` and `b`

    Parameters
    ----------
    a,b : array_like
        Sample observations. These arrays must have the same length and either
        an equivalent number of columns or be broadcastable
    ddof : int, optional
        Degrees of freedom correction in the calculation of the standard
        deviation. Default: 1
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle when input contains nan. 'propagate' returns nan,
        'raise' throws an error, 'omit' performs the calculations ignoring nan
        values. Default: 'propagate'

    Returns
    -------
    corr : float or numpy.ndarray
        Pearson's correlation coefficient between matching columns of inputs
    pval : float or numpy.ndarray
        Two-tailed p-values. Only returned if `return_pval` is True

    Notes
    -----
    If either input contains nan and nan_policy is set to 'omit', both arrays
    will be masked to omit the nan entries.
    """

    a, b, axis = _chk2_asarray(a, b, 0)
    if len(a) != len(b):
        raise ValueError('Provided arrays do not have same length')

    if a.size == 0 or b.size == 0:
        return np.nan, np.nan

    if nan_policy not in ('propagate', 'raise', 'omit'):
        raise ValueError(f'Value for nan_policy "{nan_policy}" not allowed')

    a, b = a.reshape(len(a), -1), b.reshape(len(b), -1)
    if (a.shape[1] != b.shape[1]):
        a, b = np.broadcast_arrays(a, b)

    mask = np.logical_or(np.isnan(a), np.isnan(b))
    if nan_policy == 'raise' and np.any(mask):
        raise ValueError('Input contains nan')
    elif nan_policy == 'omit':
        # avoid making copies of the data, if possible
        a = np.ma.masked_array(a, mask, copy=False, fill_value=np.nan)
        b = np.ma.masked_array(b, mask, copy=False, fill_value=np.nan)

    with np.errstate(invalid='ignore'):
        corr = (sstats.zscore(a, ddof=ddof, nan_policy=nan_policy)
                * sstats.zscore(b, ddof=ddof, nan_policy=nan_policy))

    sumfunc, n_obs = np.sum, len(a)
    if nan_policy == 'omit':
        corr = corr.filled(np.nan)
        sumfunc = np.nansum
        n_obs = np.squeeze(np.sum(np.logical_not(np.isnan(corr)), axis=0))

    corr = sumfunc(corr, axis=0) / (n_obs - 1)
    corr = np.squeeze(np.clip(corr, -1, 1)) / 1

    if return_pval:
        # taken from scipy.stats
        ab = (n_obs / 2) - 1
        prob = 2 * special.btdtr(ab, ab, 0.5 * (1 - np.abs(corr)))

        return corr, prob

    return corr



"""
Functions for computing null models
"""

__all__ = [
    'naive_nonparametric', 'alexander_bloch', 'vazquez_rodriguez', 'vasa',
    'hungarian', 'baum', 'cornblath', 'burt2018', 'burt2020', 'moran'
]

from neuromaps.nulls.nulls import (
    alexander_bloch, vazquez_rodriguez, vasa, hungarian, baum, cornblath,
    burt2018, burt2020, moran
)



# -*- coding: utf-8 -*-
"""
Functions for operating on images + surfaces
"""

import gzip
import os
from pathlib import Path
from typing import Iterable

import nibabel as nib
from nibabel.filebasedimages import ImageFileError
import numpy as np
from scipy.interpolate import griddata

PARCIGNORE = [
    'unknown', 'corpuscallosum', 'Background+FreeSurfer_Defined_Medial_Wall',
    '???', 'Unknown', 'Medial_wall', 'Medial wall', 'medial_wall'
]


def construct_surf_gii(vert, tri):
    """
    Constructs surface gifti image from `vert` and `tri`

    Parameters
    ----------
    vert : (N, 3)
        Vertices of surface mesh
    tri : (T, 3)
        Triangles comprising surface mesh

    Returns
    -------
    img : nib.gifti.GiftiImage
        Surface image
    """

    vert = nib.gifti.GiftiDataArray(vert, 'NIFTI_INTENT_POINTSET',
                                    'NIFTI_TYPE_FLOAT32',
                                    coordsys=nib.gifti.GiftiCoordSystem(3, 3))
    tri = nib.gifti.GiftiDataArray(tri, 'NIFTI_INTENT_TRIANGLE',
                                   'NIFTI_TYPE_INT32')
    img = nib.GiftiImage(darrays=[vert, tri])

    return img


def construct_shape_gii(data, names=None, intent='NIFTI_INTENT_SHAPE',
                        labels=None):
    """
    Constructs shape gifti image from `data`

    Parameters
    ----------
    data : (N[, F]) array_like
        Input data (where `F` corresponds to different features, if applicable)

    Returns
    -------
    img : nib.gifti.GiftiImage
        Shape image
    """

    intent_dtypes = {
        'NIFTI_INTENT_SHAPE': 'float32',
        'NIFTI_INTENT_LABEL': 'int32'
    }
    dtype = intent_dtypes.get(intent, 'float32')

    if data.ndim == 1:
        data = data[:, None]
    if names is not None:
        if len(names) != data.shape[1]:
            raise ValueError('Length of provided `names` does not match '
                             'number of features in `data`')
        names = [{'Name': name} for name in names]
    else:
        names = [{} for _ in range(data.shape[1])]

    labeltable = None
    if labels is not None and intent == 'NIFTI_INTENT_LABEL':
        labeltable = nib.gifti.GiftiLabelTable()
        for key, label in enumerate(labels):
            glabel = nib.gifti.GiftiLabel(key)
            glabel.label = label
            labeltable.labels.append(glabel)

    return nib.GiftiImage(darrays=[
        nib.gifti.GiftiDataArray(darr.astype(dtype), intent=intent,
                                 datatype=f'NIFTI_TYPE_{dtype.upper()}',
                                 meta=meta)
        for darr, meta in zip(data.T, names)
    ], labeltable=labeltable)


def fix_coordsys(fn, val=3):
    """
    Sets {xform,data}space of coordsys for GIFTI image `fn` to `val`

    Parameters
    ----------
    fn : str or os.PathLike
        Path to GIFTI image

    Returns
    -------
    fn : os.PathLike
        Path to GIFTI image
    """

    fn = Path(fn)
    img = nib.load(fn)
    for attr in ('dataspace', 'xformspace'):
        setattr(img.darrays[0].coordsys, attr, val)
    nib.save(img, fn)
    return fn


def load_nifti(img):
    """
    Loads nifti file `img`

    Parameters
    ----------
    img : os.PathLike or nib.Nifti1Image object
        Image to be loaded

    Returns
    -------
    img : nib.Nifti1Image
        Loaded NIFTI image
    """

    try:
        img = nib.load(img)
    except (TypeError) as err:
        msg = ('stat: path should be string, bytes, os.PathLike or integer, '
               'not Nifti1Image')
        if not str(err) == msg:
            raise err
    return img


def load_gifti(img):
    """
    Loads gifti file `img`

    Will try to gunzip `img` if gzip is detected, and will pass pre-loaded
    GiftiImage object

    Parameters
    ----------
    img : os.PathLike or nib.GiftiImage object
        Image to be loaded

    Returns
    -------
    img : nib.GiftiImage
        Loaded GIFTI images
    """

    try:
        img = nib.load(img)
    except (ImageFileError, TypeError) as err:
        # it's gzipped, so read the gzip and pipe it in
        if isinstance(err, ImageFileError) and str(err).endswith('.gii.gz"'):
            with gzip.GzipFile(img) as gz:
                img = nib.GiftiImage.from_bytes(gz.read())
        # it's not a pre-loaded GiftiImage so error out
        elif (isinstance(err, TypeError)
              and not str(err) == 'stat: path should be string, bytes, os.'
                                  'PathLike or integer, not GiftiImage'):
            raise err

    return img


def load_data(data):
    """
    Small utility function to load + stack `data` images (gifti / nifti)

    Parameters
    ----------
    data : tuple-of-str or os.PathLike or nib.GiftiImage or nib.Nifti1Image
        Data to be loaded

    Returns
    -------
    out : np.ndarray
        Loaded `data`
    """

    if (isinstance(data, (str, os.PathLike, np.ndarray))
            or not isinstance(data, Iterable)):
        data = (data,)

    try:
        out = np.hstack([load_gifti(img).agg_data() for img in data])
    except (AttributeError, TypeError):
        out = np.stack([load_nifti(img).get_fdata() for img in data], axis=3)
    except ValueError as err:
        if str(err) == 'stat: embedded null character in path':
            out = np.stack(data, axis=-1)

    return np.squeeze(out)


def obj_to_gifti(obj, fn=None):
    """
    Converts CIVET `obj` surface file to GIFTI format

    Parameters
    ----------
    obj : str or os.PathLike
        CIVET file to be converted
    fn : str or os.PathLike, None
        Output filename. If not supplied uses input `obj` filename (with
        appropriate suffix). Default: None

    Returns
    -------
    fn : os.PathLike
        Path to saved image file
    """

    from neuromaps.civet import read_civet_surf

    img = construct_surf_gii(*read_civet_surf(Path(obj)))
    if fn is None:
        fn = obj
    fn = Path(fn).resolve()
    if fn.name.endswith('.obj'):
        fn = fn.parent / fn.name.replace('.obj', '.surf.gii')
    nib.save(img, fn)

    return fn


def fssurf_to_gifti(surf, fn=None):
    """
    Converts FreeSurfer `surf` surface file to GIFTI format

    Parameters
    ----------
    obj : str or os.PathLike
        FreeSurfer surface file to be converted
    fn : str or os.PathLike, None
        Output filename. If not supplied uses input `surf` filename (with
        appropriate suffix). Default: None

    Returns
    -------
    fn : os.PathLike
        Path to saved image file
    """

    img = construct_surf_gii(*nib.freesurfer.read_geometry(Path(surf)))
    if fn is None:
        fn = surf + '.surf.gii'
    fn = Path(fn)
    nib.save(img, fn)

    return fn


def fsmorph_to_gifti(morph, fn=None, modifier=None):
    """
    Converts FreeSurfer `morph` data file to GIFTI format

    Parameters
    ----------
    obj : str or os.PathLike
        FreeSurfer morph file to be converted
    fn : str or os.PathLike, None
        Output filename. If not supplied uses input `morph` filename (with
        appropriate suffix). Default: None
    modifier : float, optional
        Scalar factor to modify (multiply) the morphometric data. Default: None

    Returns
    -------
    fn : os.PathLike
        Path to saved image file
    """

    data = nib.freesurfer.read_morph_data(Path(morph))
    if modifier is not None:
        data *= float(modifier)
    img = construct_shape_gii(data)
    if fn is None:
        fn = morph + '.shape.gii'
    fn = Path(fn)
    nib.save(img, fn)

    return fn


def interp_surface(data, src, trg, method='nearest'):
    """
    Interpolate `data` on `src` surface to `trg` surface

    Parameters
    ----------
    data : str or os.PathLike
        Path to (gifti) data file defined on `src` surface
    src : str or os.PathLike
        Path to (gifti) file defining surface of `data`
    trg : str or os.PathLike
        Path to (gifti) file defining desired output surface
    method : {'nearest', 'linear'}
        Method for interpolation. Default {'nearest'}

    Returns
    -------
    interp : np.ndarray
        Input `data` interpolated to `trg` surface
    """

    if method not in ('nearest', 'linear'):
        raise ValueError(f'Provided method {method} invalid')

    src = load_gifti(src).agg_data('NIFTI_INTENT_POINTSET')
    data = load_gifti(data).agg_data()
    if len(src) != len(data):
        raise ValueError('Provided `src` file has different number of '
                         'vertices from `data` file')
    trg = load_gifti(trg).agg_data('NIFTI_INTENT_POINTSET')

    return griddata(src, data, trg, method=method)


def vertex_areas(surface):
    """
    Calculates vertex areas from `surface` file

    Vertex area is calculated as the sum of 1/3 the area of each triangle in
    which the vertex participates

    Parameters
    ----------
    surface : str or os.PathLike
        Path to (gifti) file defining surface for which areas should be
        computed

    Returns
    -------
    areas : np.ndarray
        Vertex areas
    """

    vert, tri = load_gifti(surface).agg_data()
    vectors = np.diff(vert[tri], axis=1)
    cross = np.cross(vectors[:, 0], vectors[:, 1])
    triareas = (np.sqrt(np.sum(cross ** 2, axis=1)) * 0.5) / 3
    areas = np.bincount(tri.flatten(), weights=np.repeat(triareas, 3))

    return areas


def average_surfaces(*surfs):
    """
    Generates average surface from input `surfs`

    Parameters
    ----------
    surfs : str or os.PathLike
        Path to (gifti) surfaces to be averaged. Surfaces should be aligned!

    Returns
    -------
    average : nib.gifti.GiftiImage
        Averaged surface
    """

    n_surfs = len(surfs)
    vertices = triangles = None
    for surf in surfs:
        img = load_gifti(surf)
        vert = img.agg_data('NIFTI_INTENT_POINTSET')
        if vertices is None:
            vertices = np.zeros_like(vert)
        if triangles is None:
            triangles = img.agg_data('NIFTI_INTENT_TRIANGLE')
        vertices += vert

    vertices /= n_surfs

    return construct_surf_gii(vertices, triangles)


def _relabel(labels, minval=0, bgval=None):
    """
    Relabels `labels` so that they're consecutive

    Parameters
    ----------
    labels : (N,) array_like
        Labels to be re-labelled
    minval : int, optional
        What the new minimum value of the labels should be. Default: 0
    bgval : int, optional
        What the background value should be; the new labels will start at
        `minval` but the first value of these labels (i.e., labels == `minval`)
        will be set to `bgval`. Default: None

    Returns
    ------
    labels : (N,) np.ndarray
        New labels
    """

    labels = np.unique(labels, return_inverse=True)[-1] + minval
    if bgval is not None:
        labels[labels == minval] = bgval
    return labels


def relabel_gifti(parcellation, background=None, offset=None):
    """
    Updates GIFTI images so label IDs are consecutive across hemispheres

    Parameters
    ----------
    parcellation : (2,) tuple-of-str
        Surface label files in GIFTI format (lh.label.gii, rh.label.gii)
    background : list-of-str, optional
        If provided, a list of IDs in `parcellation` that should be set to 0
        (the presumptive background value). Other IDs will be shifted so they
        are consecutive (i.e., 0--N). If not specified will use labels in
        `neuromaps.images.PARCIGNORE`. Default: None
    offset : int, optional
        What the lowest value in `parcellation[1]` should be not including
        background value. If not specified it will be purely consecutive from
        `parcellation[0]`. Default: None

    Returns
    -------
    relabelled : (2,) tuple-of-nib.gifti.GiftiImage
        Re-labelled `parcellation` files
    """

    relabelled = tuple()
    minval = 0
    if not isinstance(parcellation, tuple):
        parcellation = (parcellation,)

    if background is None:
        background = PARCIGNORE.copy()

    for hemi in parcellation:
        # get necessary info from file
        img = load_gifti(hemi)
        data = img.agg_data().copy()
        labels = img.labeltable.labels
        lt = {v: k for k, v in img.labeltable.get_labels_as_dict().items()}

        # get rid of labels we want to drop
        if background is not None and len(labels) > 0:
            for val in background:
                idx = lt.get(val, 0)
                if idx == 0:
                    continue
                data[data == idx] = 0
                labels = [f for f in labels if f.key != idx]

        # reset labels so they're consecutive and update label keys
        data = _relabel(np.clip(data, 0, None), minval=minval, bgval=0)
        ids = np.unique(data)
        new_labels = []
        if len(labels) > 0:
            for n, i in enumerate(ids):
                lab = labels[n]
                lab.key = i
                new_labels.append(lab)
        minval = len(ids) - 1 if offset is None else int(offset) - 1

        # make new gifti image with updated information
        darr = nib.gifti.GiftiDataArray(data, intent='NIFTI_INTENT_LABEL',
                                        datatype='NIFTI_TYPE_INT32')
        labeltable = nib.gifti.GiftiLabelTable()
        labeltable.labels = new_labels
        img = nib.GiftiImage(darrays=[darr], labeltable=labeltable)
        relabelled += (img,)

    return relabelled


def annot_to_gifti(parcellation, background=None):
    """
    Converts FreeSurfer-style annotation `parcellation` files to GIFTI images

    Parameters
    ----------
    parcellation : tuple of str or os.PathLike
        Paths to surface annotation files (.annot)
    background : list-of-str, optional
        If provided, a list of IDs in `parcellation` that should be set to 0
        (the presumptive background value). Other IDs will be shifted so they
        are consecutive (i.e., 0--N). If not specified will use labels in
        `neuromaps.images.PARCIGNORE`. Default: None

    Returns
    -------
    gifti : tuple-of-nib.GiftiImage
        Converted GIFTI images
    """

    if not isinstance(parcellation, tuple):
        parcellation = (parcellation,)

    gifti = tuple()
    for atlas in parcellation:
        labels, ctab, names = nib.freesurfer.read_annot(atlas)

        darr = nib.gifti.GiftiDataArray(labels, intent='NIFTI_INTENT_LABEL',
                                        datatype='NIFTI_TYPE_INT32')
        labeltable = nib.gifti.GiftiLabelTable()
        for key, label in enumerate(names):
            (r, g, b), a = (ctab[key, :3] / 255), (1.0 if key != 0 else 0.0)
            glabel = nib.gifti.GiftiLabel(key, r, g, b, a)
            glabel.label = label.decode()
            labeltable.labels.append(glabel)

        gifti += (nib.GiftiImage(darrays=[darr], labeltable=labeltable),)

    return relabel_gifti(gifti, background=background)


def dlabel_to_gifti(parcellation):
    """
    Converts CIFTI dlabel file to GIFTI images

    Parameters
    ----------
    parcellation : str or os.PathLike
        Path to CIFTI parcellation file (.dlabel.nii)

    Returns
    -------
    gifti : tuple-of-nib.GiftiImage
        Converted GIFTI images
    """

    structures = ('CORTEX_LEFT', 'CORTEX_RIGHT')

    dlabel = nib.load(parcellation)
    parcdata = np.asarray(dlabel.get_fdata(), dtype='int32').squeeze()

    gifti = tuple()
    label_dict = dlabel.header.get_axis(index=0).label[0]
    for bm in dlabel.header.get_index_map(1).brain_models:
        structure = bm.brain_structure
        if structure.startswith('CIFTI_STRUCTURE_'):
            structure = structure[16:]
        if structure not in structures:
            continue
        labels = np.zeros(bm.surface_number_of_vertices, dtype='int32')
        idx = np.asarray(bm.vertex_indices)
        slicer = slice(bm.index_offset, bm.index_offset + bm.index_count)
        labels[idx] = parcdata[slicer]

        darr = nib.gifti.GiftiDataArray(labels, intent='NIFTI_INTENT_LABEL',
                                        datatype='NIFTI_TYPE_INT32')
        labeltable = nib.gifti.GiftiLabelTable()
        for key, (label, (r, g, b, a)) in label_dict.items():
            if key not in labels:
                continue
            glabel = nib.gifti.GiftiLabel(key, r, g, b, a)
            glabel.label = label
            labeltable.labels.append(glabel)
        gifti += (nib.GiftiImage(darrays=[darr], labeltable=labeltable),)

    return gifti


def minc_to_nifti(img, fn=None):
    """
    Converts MINC `img` to NIfTI format (and re-orients to RAS)

    Parameters
    ----------
    img : str or os.PathLike
        Path to MINC file to be converted
    fn : str or os.PathLike, optional
        Filepath to where converted NIfTI image should be stored. If not
        supplied the converted image is not saved to disk and is returned.
        Default: None

    Returns
    -------
    out : nib.Nifti1Image or os.PathLike
        Converted image (if `fn` is None) or path to saved file on disk
    """

    mnc = nib.load(img)
    nifti = nib.Nifti1Image(np.asarray(mnc.dataobj), mnc.affine)

    # re-orient nifti image RAS
    orig_ornt = nib.io_orientation(nifti.affine)
    targ_ornt = nib.orientations.axcodes2ornt('RAS')
    transform = nib.orientations.ornt_transform(orig_ornt, targ_ornt)
    nifti = nifti.as_reoriented(transform)

    # save file (if desired)
    if fn is not None:
        fn = Path(fn).resolve()
        if fn.name.endswith('.mnc'):
            fn = fn.parent / fn.name.replace('.mnc', '.nii.gz')
        nib.save(nifti, fn)
        return fn
    return nifti



# -*- coding: utf-8 -*-
"""
For testing neuromaps.resampling functionality
"""

import pytest

from neuromaps import resampling


@pytest.mark.xfail
def test__estimate_density():
    assert False


@pytest.mark.xfail
@pytest.mark.workbench
def test_downsample_only():
    assert False


@pytest.mark.xfail
@pytest.mark.workbench
def test_transform_to_src():
    assert False


@pytest.mark.xfail
@pytest.mark.workbench
def test_transform_to_trg():
    assert False


@pytest.mark.xfail
@pytest.mark.workbench
def test_transform_to_alt():
    assert False


@pytest.mark.xfail
def test_mni_transform():
    assert False


def test__check_altspec():
    spec = ('fsaverage', '10k')
    assert resampling._check_altspec(spec) == spec

    for spec in (None, ('fsaverage',), ('fsaverage', '100k')):
        with pytest.raises(ValueError):
            resampling._check_altspec(spec)


@pytest.mark.xfail
@pytest.mark.workbench
def test_resample_images():
    assert False



# -*- coding: utf-8 -*-
"""
For testing neuromaps.parcellate functionality
"""

import pytest


@pytest.mark.xfail
def test__gifti_to_array():
    assert False


@pytest.mark.xfail
def test__array_to_gifti():
    assert False


@pytest.mark.xfail
def test_Parcellater():
    assert False



# -*- coding: utf-8 -*-
"""
CARET files I/O
"""

from io import BytesIO
import re
import struct

import nibabel as nib
import numpy as np


def read_surface_shape(fn):
    """
    Reads surface_shape CARET file

    Parameters
    ----------
    fn : str
        Filepath to surface_shape file

    Returns
    -------
    names : (F,) list-of-str
        Names of columns in `data`
    data : (N, F) np.ndarray
        Data from `fn`
    """

    with open(fn, 'rb') as src:
        surfshape = src.read()
    msg = b'BEGIN-DATA\n'
    offset = surfshape.find(msg) + len(msg)
    header, data = surfshape[:offset].decode(), surfshape[offset:]
    n_nodes = int(re.search(r'tag-number-of-nodes (\d+)', header).group(1))
    n_cols = int(re.search(r'tag-number-of-columns (\d+)', header).group(1))
    names = re.findall(r'tag-column-name \d+ ([\S\s][^\n]*)', header)
    encoding = re.search(r'encoding (\S+)', header).group(1)
    if encoding == 'ASCII':
        data = np.loadtxt(BytesIO(data))
    else:
        data = np.asarray(struct.unpack('>' + ('f' * n_cols * n_nodes), data))
        data = data.reshape(-1, n_cols)
    return names, data


def read_coords(fn):
    """
    Reads coords CARET file

    Parameters
    ----------
    fn : str
        Filepath to coords file

    Returns
    -------
    coords : (N, 3) np.ndarray
        Coordinates from `fn`
    """

    with open(fn, 'rb') as src:
        coords = src.read()
    msg = b'EndHeader\n'
    offset = coords.find(msg) + len(msg)
    _, data = coords[:offset].decode(), coords[offset:]
    n_nodes, = struct.unpack('>i', data[:4])
    data = np.asarray(struct.unpack('>' + 'fff' * n_nodes, data[4:]))
    data = data.reshape(-1, 3)
    return data


def read_topo(fn):
    """
    Reads topo CARET file

    Parameters
    ----------
    fn : str
        Filepath to topo file

    Returns
    -------
    topo : (T, 3) np.ndarray
        Topology from `fn`
    """

    with open(fn, 'rb') as src:
        topo = src.read()
    msg = b'tag-version 1\n'
    if msg not in topo:
        raise ValueError('Cannot load topology data in old format')
    offset = topo.find(msg) + len(msg)
    header, data = topo[:offset].decode(), topo[offset:]
    encoding = re.search(r'encoding (\S+)', header).group(1)
    if encoding == "ASCII":
        data = np.loadtxt(BytesIO('\n'.join(data.split('\n')[1:])))
    else:
        n_nodes, = struct.unpack('>i', data[:4])
        data = np.asarray(struct.unpack('>' + 'iii' * n_nodes, data[4:]))
        data = data.reshape(-1, 3)
    return data


def read_deform_map(fn):
    """
    Reads deform_map CARET file

    Parameters
    ----------
    fn : str
        Filepath to deform_map file

    Returns
    -------
    nodes : (N, 3) np.ndarray
        Neighboring nodes for surface
    barycentric : (N, 3) np.ndarray
        Barycentric coordinates for surface
    """

    with open(fn, 'rb') as src:
        deform = src.read()
    msg = b'DATA-START\n'
    offset = deform.find(msg) + len(msg)
    header, data = deform[:offset].decode(), deform[offset:]
    encoding = re.search(r'encoding (\S+)', header).group(1)
    if encoding == 'ASCII':
        data = np.loadtxt(BytesIO(data))
    else:
        n_nodes, = struct.unpack('>i', data[:4])
        data = np.asarray(struct.unpack('>' + 'iiifff' * n_nodes, data[4:]))
        data = data.reshape(-1, 6)
    nodes, barycentric = data[:, :3].astype(int), data[:, 3:]
    return nodes, barycentric


def apply_deform_map(source, deformation, method='nearest'):
    """
    Applies `deformation` map to `source` brainmap

    Parameters
    ----------
    source : str or os.PathLike
        Path to (gifti) file, to be deformed
    deformation : str or os.PathLike
        Path to deformation map
    methods : {'nearest'}, optional
        Method for applying `deformation` to `source`. Currently only 'nearest'
        is supported. Default: 'nearest'

    Returns
    -------
    projected : (N,) np.ndarray
        Data from `source` projected to surface specified in `deformation`
    """

    methods = {'nearest'}
    if method not in methods:
        raise ValueError(f'Invalid method {method}. Must be one of {methods}')

    nodes, bary = read_deform_map(deformation)
    data = nib.load(source).agg_data()

    if method == 'average':  # FIXME: this is wrong.
        projected = np.sum(data[nodes] * bary, axis=1)
    elif method == 'nearest':
        projected = data[nodes[:, 0]]

    return projected



# -*- coding: utf-8 -*-
"""
For testing neuromaps.civet functionality
"""

import pytest


@pytest.mark.xfail
def test_read_civet_surf():
    assert False


@pytest.mark.xfail
def test_read_surfmap():
    assert False


@pytest.mark.xfail
def test_resample_surface_map():
    assert False



__all__ = ['resample_images', 'compare_images']

from neuromaps.resampling import resample_images
from neuromaps.stats import compare_images

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
"""
Functions for working with triangle meshes + surfaces
"""

from joblib import Parallel, delayed
import numpy as np
from scipy import ndimage, sparse

from neuromaps.images import load_gifti, relabel_gifti, PARCIGNORE


def point_in_triangle(point, triangle, return_pdist=True):
    """
    Checks whether `point` falls inside `triangle`

    Parameters
    ----------
    point : (3,) array_like
        Coordinates of point
    triangle (3, 3) array_like
        Coordinates of triangle
    return_pdist : bool, optional
        Whether to return planar distance (see outputs). Default: True

    Returns
    -------
    inside : bool
        Whether `point` is inside triangle
    pdist : float
        The approximate distance of the point to the plane of the triangle.
        Only returned if `return_pdist` is True
    """

    A, B, C = triangle
    v0 = C - A
    v1 = B - A
    v2 = point - A

    dot00 = np.dot(v0, v0)
    dot01 = np.dot(v0, v1)
    dot02 = np.dot(v0, v2)
    dot11 = np.dot(v1, v1)
    dot12 = np.dot(v1, v2)

    denom = 1 / (dot00 * dot11 - dot01 * dot01)
    u = (dot11 * dot02 - dot01 * dot12) * denom
    v = (dot00 * dot12 - dot01 * dot02) * denom
    inside = (u >= 0) and (v >= 0) and (u + v < 1)

    if return_pdist:
        return inside, np.abs(v2 @ np.cross(v1, v0))

    return inside


def which_triangle(point, triangles):
    """
    Determines which of `triangles` the provided `point` falls inside

    Parameters
    ----------
    point : (3,) array_like
        Coordinates of point
    triangles : (N, 3, 3) array_like
        Coordinates of `N` triangles to check

    Returns
    -------
    idx : int
        Index of `triangles` that `point` is inside of. If `point` does not
        fall within any of `triangles` then this will be None
    """

    idx, planar = None, np.inf
    for n, tri in enumerate(triangles):
        inside, pdist = point_in_triangle(point, tri)
        if pdist < planar and inside:
            idx, planar = n, pdist

    return idx


def _get_edges(faces):
    """
    Gets set of edges from `faces`

    Parameters
    ----------
    faces : (F, 3) array_like
        Set of indices creating triangular faces of a mesh

    Returns
    -------
    edges : (F*3, 2) array_like
        All edges in `faces`
    """

    faces = np.asarray(faces)
    edges = np.sort(faces[:, [0, 1, 1, 2, 2, 0]].reshape((-1, 2)), axis=1)

    return edges


def get_shared_triangles(faces):
    """
    Returns dictionary of triangles sharing edges from `faces`

    Parameters
    ----------
    faces : (N, 3)
        Triangles comprising mesh

    Returns
    -------
    shared : dict
        Where keys are len-2 tuple of vertex ids for the shared edge and values
        are the triangles that have this shared edge.
    """

    # first generate the list of edges for the provided faces and the
    # index for which face the edge is from (which is just the index of the
    # face repeated thrice, since each face generates three direct edges)
    edges = _get_edges(faces)
    edges_face = np.repeat(np.arange(len(faces)), 3)

    # every edge appears twice in a watertight surface, so we'll first get the
    # indices for each duplicate edge in `edges` (this should, assuming all
    # goes well, have rows equal to len(edges) // 2)
    order = np.lexsort(edges.T[::-1])
    edges_sorted = edges[order]
    dupe = np.any(edges_sorted[1:] != edges_sorted[:-1], axis=1)
    dupe_idx = np.append(0, np.nonzero(dupe)[0] + 1)
    start_ok = np.diff(np.concatenate((dupe_idx, [len(edges_sorted)]))) == 2
    groups = np.tile(dupe_idx[start_ok].reshape(-1, 1), 2)
    edge_groups = order[groups + np.arange(2)]

    # now, get the indices of the faces that participate in these duplicate
    # edges, as well as the edges themselves
    adjacency = edges_face[edge_groups]
    nondegenerate = adjacency[:, 0] != adjacency[:, 1]
    adjacency = np.sort(adjacency[nondegenerate], axis=1)
    adjacency_edges = edges[edge_groups[:, 0][nondegenerate]]

    # the non-shared vertex index is the same shape as adjacency, holding
    # vertex indices vs face indices
    indirect_edges = np.zeros(adjacency.shape, dtype=np.int32) - 1

    # loop through the two columns of adjacency
    for i, fid in enumerate(adjacency.T):
        # faces from the current column of adjacency
        face = faces[fid]
        # get index of vertex not included in shared edge
        unshared = np.logical_not(np.logical_or(
            face == adjacency_edges[:, 0].reshape(-1, 1),
            face == adjacency_edges[:, 1].reshape(-1, 1)))
        # each row should have one "uncontained" vertex; ignore degenerates
        row_ok = unshared.sum(axis=1) == 1
        unshared[~row_ok, :] = False
        indirect_edges[row_ok, i] = face[unshared]

    # get vertex coordinates of triangles pairs with shared edges, ordered
    # such that the non-shared vertex is always _last_ among the trio
    shared = np.sort(face[np.logical_not(unshared)].reshape(-1, 1, 2), axis=-1)
    shared = np.repeat(shared, 2, axis=1)
    triangles = np.concatenate((shared, indirect_edges[..., None]), axis=-1)

    return dict(zip(map(tuple, adjacency_edges), triangles))


def get_direct_edges(vertices, faces):
    """
    Gets (unique) direct edges and weights in mesh describes by inputs.

    Parameters
    ----------
    vertices : (N, 3) array_like
        Coordinates of `vertices` comprising mesh with `faces`
    faces : (F, 3) array_like
        Indices of `vertices` that compose triangular faces of mesh

    Returns
    -------
    edges : (E, 2) array_like
        Indices of `vertices` comprising direct edges (without duplicates)
    weights : (E, 1) array_like
        Distances between `edges`
    """

    edges = np.unique(_get_edges(faces), axis=0)
    weights = np.linalg.norm(np.diff(vertices[edges], axis=1), axis=-1)
    return edges, weights.squeeze()


def get_indirect_edges(vertices, faces):
    """
    Gets indirect edges and weights in mesh described by inputs

    Indirect edges are between two vertices that participate in faces sharing
    an edge

    Parameters
    ----------
    vertices : (N, 3) array_like
        Coordinates of `vertices` comprising mesh with `faces`
    faces : (F, 3) array_like
        Indices of `vertices` that compose triangular faces of mesh

    Returns
    -------
    edges : (E, 2) array_like
        Indices of `vertices` comprising indirect edges (without duplicates)
    weights : (E, 1) array_like
        Distances between `edges` on surface

    References
    ----------
    https://github.com/mikedh/trimesh (MIT licensed)
    """

    triangles = np.stack(list(get_shared_triangles(faces).values()), axis=0)
    indirect_edges = triangles[..., -1]

    # `A.shape`: (3, N, 2) corresponding to (xyz coords, edges, triangle pairs)
    A, B, V = vertices[triangles].transpose(2, 3, 0, 1)

    # calculate the xyz coordinates of the foot of each triangle, where the
    # base is the shared edge
    # that is, we're trying to calculate F in the equation `VF = VB - (w * BA)`
    # where `VF`, `VB`, and `BA` are vectors, and `w = (AB * VB) / (AB ** 2)`
    w = (np.sum((A - B) * (V - B), axis=0, keepdims=True)
         / np.sum((A - B) ** 2, axis=0, keepdims=True))
    feet = B - (w * (B - A))
    # calculate coordinates of midpoint b/w the feet of each pair of triangles
    midpoints = (np.sum(feet.transpose(1, 2, 0), axis=1) / 2)[:, None]
    # calculate Euclidean distance between non-shared vertices and midpoints
    # and add distances together for each pair of triangles
    norms = np.linalg.norm(vertices[indirect_edges] - midpoints, axis=-1)
    weights = np.sum(norms, axis=-1)

    # NOTE: weights won't be perfectly accurate for a small subset of triangle
    # pairs where either triangle has angle >90 along the shared edge. in these
    # the midpoint lies _outside_ the shared edge, so neighboring triangles
    # would need to be taken into account. that said, this occurs in only a
    # minority of cases and the difference tends to be in the ~0.001 mm range
    return indirect_edges, weights


def make_surf_graph(vertices, faces, mask=None):
    """
    Constructs adjacency graph from `surf`.

    Parameters
    ----------
    vertices : (N, 3) array_like
        Coordinates of `vertices` comprising mesh with `faces`
    faces : (F, 3) array_like
        Indices of `vertices` that compose triangular faces of mesh
    mask : (N,) array_like, optional (default None)
        Boolean mask indicating which vertices should be removed from generated
        graph. If not supplied, all vertices are used.

    Returns
    -------
    graph : scipy.sparse.csr_matrix
        Sparse matrix representing graph of `vertices` and `faces`

    Raises
    ------
    ValueError
        Inconsistent number of vertices in `mask` and `vertices`
    """

    if mask is not None and len(mask) != len(vertices):
        raise ValueError('Supplied `mask` array has different number of '
                         'vertices than supplied `vertices`.')

    # get all (direct + indirect) edges from surface
    direct_edges, direct_weights = get_direct_edges(vertices, faces)
    indirect_edges, indirect_weights = get_indirect_edges(vertices, faces)
    edges = np.row_stack((direct_edges, indirect_edges))
    weights = np.hstack((direct_weights, indirect_weights))

    # remove edges that include a vertex in `mask`
    if mask is not None:
        idx, = np.where(mask)
        mask = ~np.any(np.isin(edges, idx), axis=1)
        edges, weights = edges[mask], weights[mask]

    # construct our graph on which to calculate shortest paths
    return sparse.csr_matrix((np.squeeze(weights), (edges[:, 0], edges[:, 1])),
                             shape=(len(vertices), len(vertices)))


def _get_graph_distance(vertex, graph, labels=None):
    """
    Gets surface distance of `vertex` to all other vertices in `graph`

    Parameters
    ----------
    vertex : int
        Index of vertex for which to calculate surface distance
    graph : array_like
        Graph along which to calculate shortest path distances
    labels : array_like, optional
        Labels indicating parcel to which each vertex belongs. If provided,
        distances will be averaged within distinct labels

    Returns
    -------
    dist : (N,) numpy.ndarray
        Distance of `vertex` to all other vertices in `graph` (or to all
        parcels in `labels`, if provided)
    """

    dist = sparse.csgraph.dijkstra(graph, directed=False, indices=[vertex])

    if labels is not None:
        dist = ndimage.mean(input=np.delete(dist, vertex),
                            labels=np.delete(labels, vertex),
                            index=np.unique(labels))

    return dist.astype('float32')


def get_surface_distance(surface, parcellation=None, medial=None,
                         medial_labels=None, drop=None, n_proc=1):
    """
    Calculates surface distance for vertices in `surface`

    Parameters
    ----------
    surface : str or os.PathLike
        Path to surface file on which to calculate distance
    parcellation : str or os.PathLike, optional
        Path to file with parcel labels for provided `surface`. If provided
        will calculate parcel-parcel distances instead of vertex distances,
        where parcel-parcel distance is the average distance between all
        constituent vertices in two parcels. Default: None
    medial : str or os.PathLike, optional
        Path to file indicating which vertices correspond to the medial wall
        (0 indicates medial wall). If provided will prohibit calculation of
        surface distance along the medial wall. Superseded by `medial_labels`
        if both are provided. Default: None
    medial_labels : list of str, optional
        List of parcel names that comprise the medial wall and through which
        travel should be disallowed. Only valid if `parcellation` is provided;
        supersedes `medial` if both are provided. Default: None
    drop : list of str, optional
        List of parcel names that should be dropped from the final distance
        matrix (if `parcellation` is provided). If not specified, will ignore
        parcels commonly used to reference the medial wall (e.g., 'unknown',
        'corpuscallosum', '???', 'Background+FreeSurfer_Defined_Medial_Wall').
        Default: None
    n_proc : int, optional
        Number of processors to use for parallelizing distance calculation. If
        negative, will use max available processors plus 1 minus the specified
        number. Default: 1 (no parallelization)

    Returns
    -------
    distance : (N, N) numpy.ndarray
        Surface distance between vertices/parcels on `surface`
    """

    if drop is None:
        drop = PARCIGNORE

    if medial_labels is not None:
        if isinstance(medial_labels, str):
            medial_labels = [medial_labels]
        drop = set(drop + list(medial_labels))

    vert, faces = load_gifti(surface).agg_data()
    n_vert = vert.shape[0]
    labels, mask = None, np.zeros(n_vert, dtype=bool)

    # get data from parcellation / medial wall files if provided
    if medial is not None:
        mask = np.logical_not(load_gifti(medial).agg_data().astype(bool))
    if parcellation is not None:
        parcellation, = relabel_gifti(parcellation, background=drop)
        labels = load_gifti(parcellation).agg_data()
        mask[labels == 0] = True

    # calculate distance from each vertex to all other vertices
    graph = make_surf_graph(vert, faces, mask=mask)
    dist = np.row_stack(Parallel(n_jobs=n_proc, max_nbytes=None)(
        delayed(_get_graph_distance)(n, graph, labels) for n in range(n_vert)
    ))

    # average distance for all vertices within a parcel + set diagonal to 0
    if labels is not None:
        dist = np.row_stack([
            dist[labels == lab].mean(axis=0) for lab in np.unique(labels)
        ])
        dist[np.diag_indices_from(dist)] = 0
        dist = dist[1:, 1:]

    # remove distances for parcels that we aren't interested in
    return dist


def _geodesic_parcel_centroid(vertices, faces, inds):
    """
    Calculates parcel centroids based on surface distance

    Parameters
    ----------
    vertices : (N, 3)
        Coordinates of vertices defining surface
    faces : (F, 3)
        Triangular faces defining surface
    inds : (R,)
        Indices of `vertices` that belong to parcel

    Returns
    --------
    roi : (3,) numpy.ndarray
        Vertex corresponding to centroid of parcel
    """

    mask = np.ones(len(vertices), dtype=bool)
    mask[inds] = False
    mat = make_surf_graph(vertices, faces, mask=mask)
    paths = sparse.csgraph.dijkstra(mat, directed=False, indices=inds)[:, inds]

    # the selected vertex is the one with the minimum average shortest path
    # to the other vertices in the parcel
    roi = vertices[inds[paths.mean(axis=1).argmin()]]

    return roi



# -*- coding: utf-8 -*-
"""
For testing neuromaps.datasets._osf functionality
"""

from pkg_resources import resource_filename

import pytest

from neuromaps.datasets import _osf


@pytest.mark.xfail
def test_parse_filename():
    assert False


@pytest.mark.xfail
def test_parse_fname_list():
    assert False


def test_parse_json():
    osf = resource_filename('neuromaps', 'datasets/data/osf.json')
    out = _osf.parse_json(osf)
    assert isinstance(out, list) and all(isinstance(i, dict) for i in out)


@pytest.mark.xfail
def test_write_json():
    assert False


@pytest.mark.xfail
def test_complete_json():
    assert False


@pytest.mark.xfail
def test_check_missing_keys():
    assert False


@pytest.mark.xfail
def test_generate_auto_keys():
    assert False


@pytest.mark.xfail
def test_clean_minimal_keys():
    assert False


@pytest.mark.xfail
def test_get_url():
    assert False


@pytest.mark.xfail
def test_generate_release_json():
    assert False



# -*- coding: utf-8 -*-
"""
Functions for fetching datasets (from the internet, if necessary)
"""

from collections import namedtuple
import os
from pathlib import Path

from nilearn.datasets.utils import _fetch_files
from sklearn.utils import Bunch

from neuromaps.datasets.utils import get_data_dir, get_dataset_info

SURFACE = namedtuple('Surface', ('L', 'R'))
ALIAS = dict(
    fslr='fsLR', fsavg='fsaverage', mni152='MNI152', mni='MNI152',
    FSLR='fsLR', CIVET='civet'
)
DENSITIES = dict(
    civet=['41k', '164k'],
    fsaverage=['3k', '10k', '41k', '164k'],
    fsLR=['4k', '8k', '32k', '164k'],
    MNI152=['1mm', '2mm', '3mm'],
)


_atlas_docs = dict(
    url="""\
url : str, optional
    URL from which to download data. Default: None\
""",
    data_dir="""\
data_dir : str, optional
    Path to use as data directory. If not specified, will check for
    environmental variable 'neuromaps_DATA'; if that is not set, will
    use `~/neuromaps-data` instead. Default: None\
""",
    verbose="""\
verbose : int, optional
    Modifies verbosity of download, where higher numbers mean more updates.
    Default: 1\
""",
    genericatlas="""\
atlas : dict
    Dictionary where keys are atlas types and values are atlas files\
""",
    surfatlas="""\
atlas : dict
    Dictionary where keys are atlas types and values are tuples of atlas
    files (L/R hemisphere)\
"""
)


def _sanitize_atlas(atlas):
    """ Checks for aliases of `atlas` and confirms valid input
    """
    atlas = ALIAS.get(atlas, atlas)
    if atlas not in DENSITIES:
        raise ValueError(f'Invalid atlas: {atlas}.')
    return atlas


def _bunch_outputs(keys, values, surface=True):
    """ Groups `values` together (L/R) if `surface` and zips with `keys`
    """
    if surface:
        values = [SURFACE(*values[i:i + 2]) for i in range(0, len(values), 2)]
    return Bunch(**dict(zip(keys, values)))


def _fetch_atlas(atlas, density, keys, url=None, data_dir=None, verbose=1):
    """ Helper function to get requested `atlas`
    """

    atlas = _sanitize_atlas(atlas)
    densities = DENSITIES[atlas]
    if density not in densities:
        raise ValueError(f'Invalid density: {density}. Must be one of '
                         f'{densities}')

    data_dir = get_data_dir(data_dir=data_dir)
    info = get_dataset_info(atlas)[density]
    if url is None:
        url = info['url']
    opts = {
        'uncompress': True,
        'md5sum': info['md5'],
        'move': f'{atlas}{density}.tar.gz'
    }

    if atlas == 'MNI152':
        filenames = [
            f'tpl-MNI152NLin2009cAsym_res-{density}{suff}.nii.gz'
            for suff in ('_T1w', '_T2w', '_PD', '_desc-brain_mask',
                         '_label-csf_probseg', '_label-gm_probseg',
                         '_label-wm_probseg')
        ]
    else:
        filenames = [
            'tpl-{}_den-{}_hemi-{}_{}.surf.gii'
            .format(atlas, density, hemi, surf)
            for surf in keys
            for hemi in ('L', 'R')
        ] + [
            'tpl-{}_den-{}_hemi-{}_desc-{}.gii'
            .format(atlas, density, hemi, desc)
            for desc in ('nomedialwall_dparc.label',
                         'sulc_midthickness.shape',
                         'vaavg_midthickness.shape')
            for hemi in ('L', 'R')
        ]
        keys += ['medial', 'sulc', 'vaavg']

    filenames = [os.path.join('atlases', atlas, fn) for fn in filenames]
    data = [
        Path(fn) for fn in
        _fetch_files(data_dir, files=[(f, url, opts) for f in filenames],
                     verbose=verbose)
    ]

    return _bunch_outputs(keys, data, atlas != 'MNI152')


def fetch_civet(density='41k', url=None, data_dir=None, verbose=1):
    keys = ['white', 'midthickness', 'inflated', 'veryinflated', 'sphere']
    return _fetch_atlas(
        'civet', density, keys, url=url, data_dir=data_dir, verbose=verbose
    )


fetch_civet.__doc__ = """
Fetches CIVET surface atlas

Parameters
----------
density : {{'{densities}'}}, optional
    Density of CIVET atlas to fetch. Default: '41k'
{url}
{data_dir}
{verbose}

Returns
-------
{surfatlas}
""".format(**_atlas_docs, densities="', '".join(DENSITIES['civet']))


def fetch_fsaverage(density='41k', url=None, data_dir=None, verbose=1):
    keys = ['white', 'pial', 'inflated', 'sphere']
    return _fetch_atlas(
        'fsaverage', density, keys, url=url, data_dir=data_dir, verbose=verbose
    )


fetch_fsaverage.__doc__ = """
Fetches fsaverage surface atlas

Parameters
----------
density : {{'{densities}'}}, optional
    Density of fsaverage atlas to fetch. Default: '41k'
{url}
{data_dir}
{verbose}

Returns
-------
{surfatlas}
""".format(**_atlas_docs, densities="', '".join(DENSITIES['fsaverage']))


def fetch_fslr(density='32k', url=None, data_dir=None, verbose=1):
    keys = ['midthickness', 'inflated', 'veryinflated', 'sphere']
    if density in ('4k', '8k'):
        keys.remove('veryinflated')
    return _fetch_atlas(
        'fsLR', density, keys, url=url, data_dir=data_dir, verbose=verbose
    )


fetch_fslr.__doc__ = """
Fetches fsLR surface atlas

Parameters
----------
density : {{'{densities}'}}, optional
    Density of fsLR atlas to fetch. Default: '32k'
{url}
{data_dir}
{verbose}

Returns
-------
{surfatlas}
""".format(**_atlas_docs, densities="', '".join(DENSITIES['fsLR']))


def fetch_mni152(density='1mm', url=None, data_dir=None, verbose=1):
    keys = ['T1w', 'T2w', 'PD', 'brainmask', 'CSF', 'GM', 'WM']
    return _fetch_atlas(
        'MNI152', density, keys, url=url, data_dir=data_dir, verbose=verbose
    )


fetch_mni152.__doc__ = """
Fetches MNI152 atlas

Parameters
----------
density : {{'{densities}'}}, optional
    Resolution of MNI152 atlas to fetch. Default: '1mm'
{url}
{data_dir}
{verbose}

Returns
-------
{genericatlas}
""".format(**_atlas_docs, densities="', '".join(DENSITIES['MNI152']))


def fetch_regfusion(atlas, url=None, data_dir=None, verbose=1):
    atlas = _sanitize_atlas(atlas)
    densities = DENSITIES[atlas].copy()
    invalid = dict(civet=('164k',), fsLR=('4k', '8k'))
    for remove in invalid.get(atlas, []):
        densities.remove(remove)

    data_dir = get_data_dir(data_dir=data_dir)
    info = get_dataset_info('regfusion')
    if url is None:
        url = info['url']
    opts = {
        'uncompress': True,
        'md5sum': info['md5'],
        'move': 'regfusion.tar.gz'
    }

    filenames = [
        'tpl-MNI152_space-{}_den-{}_hemi-{}_regfusion.txt'
        .format(atlas, density, hemi)
        for density in densities
        for hemi in ['L', 'R']
    ]
    filenames = [os.path.join('atlases', 'regfusion', fn) for fn in filenames]

    data = [
        Path(fn) for fn in
        _fetch_files(data_dir, files=[(f, url, opts) for f in filenames],
                     verbose=verbose)
    ]
    return _bunch_outputs(densities, data)


fetch_regfusion.__doc__ = """
Fetches regfusion inputs for mapping MNI152 to specified surface `atlas`

Parameters
----------
atlas : {{'civet', 'fsaverage', 'fsLR'}}
    Atlas to fetch
{url}
{data_dir}
{verbose}

Returns
-------
regfusion : dict
    Dictionary where keys are surface densities and values are regfusion inputs
""".format(**_atlas_docs)


def fetch_atlas(atlas, density, url=None, data_dir=None, verbose=1):
    atlas = _sanitize_atlas(atlas)
    fetcher = globals()[f'fetch_{atlas.lower()}']
    return fetcher(density, url=url, data_dir=data_dir, verbose=verbose)


fetch_atlas.__doc__ = """
Fetches specified `atlas` and `density`

Parameters
----------
atlas : {{'{atlases}'}}
    Atlas to fetch
density : str
    Density (or resolution) of `atlas`. Must be valid for provided `atlas`
{url}
{data_dir}
{verbose}

Returns
-------
{genericatlas}
""".format(**_atlas_docs, atlases="', '".join(DENSITIES.keys()))


def fetch_all_atlases(data_dir=None, verbose=1):
    atlases = {'regfusion': {}}
    for key, resolutions in DENSITIES.items():
        atlases[key] = {}
        for res in resolutions:
            atlases[key][res] = \
                fetch_atlas(key, res, data_dir=data_dir, verbose=verbose)
        if key != 'MNI152':
            atlases['regfusion'][key] = \
                fetch_regfusion(key, data_dir=data_dir, verbose=verbose)

    return atlases


fetch_all_atlases.__doc__ = """
Fetches (and caches) all available atlases

Parameters
----------
{data_dir}
{verbose}

Returns
-------
atlases : dict
    Nested dictionaries containing all available atlases
"""


def get_atlas_dir(atlas, data_dir=None):
    try:
        atlas = _sanitize_atlas(atlas)
    except ValueError as err:
        if atlas != 'regfusion':
            raise err
    return Path(get_data_dir(data_dir=data_dir)) / 'atlases' / atlas


get_atlas_dir.__doc__ = """
Returns filepath to specified `atlas`

Parameters
----------
atlas : str
    Atlas for which filepath should be returned
{data_dir}

Returns
-------
atlas_dir : os.PathLike
    Full filepath to `atlas` directory

Raises
------
ValueError
    If provided `atlas` is not valid
""".format(**_atlas_docs)



# -*- coding: utf-8 -*-
"""
Implementation of surrogate map generation as in Burt et al., 2018, Nat Neuro
"""

import warnings

import numpy as np
from scipy.optimize import least_squares
from scipy import sparse as ssp
from scipy.stats import boxcox


def _make_weight_matrix(x, d0):
    """
    Constructs weight matrix from distance matrix + autocorrelation estimate

    Parameters
    ----------
    x : array_like
        Distance matrix
    d0 : float
        Estimate of spatial scale of autocorrelation

    Returns
    -------
    W : numpy.ndarray
        Weight matrix
    """

    with np.errstate(over='ignore'):
        weight = np.exp(-x / d0) * np.logical_not(np.eye(len(x), dtype=bool))

    with np.errstate(invalid='ignore'):
        return weight / np.sum(weight, axis=1)


def estimate_rho_d0(x, y, rho=None, d0=None):
    """
    Uses a least-squares fit to estimate `rho` and `d0`

    Parameters
    ----------
    x : array_like
        Distance matrix
    y : array_like
        Dependent brain-imaging variable; all values must be positive in order
        for successful Box-Cox transformation
    rho : float, optional
        Initial guess for rho parameter. Default: 1.0
    d0 : float, optional
        Initial guess for d0 (spatial scale of autocorrelation) parameter.
        Default: 1.0

    Returns
    -------
    rho_hat : float
        Estimate of `rho` based on least-squares fit between `x` and `y`
    d0_hat : float
        Estimate of `d0` based on least-squares fit between `x` and `y`
    """

    def _estimate(parameters, x, y):
        rho, d0 = parameters
        y_hat = rho * (_make_weight_matrix(x, d0) @ y)
        return y - y_hat

    if rho is None:
        rho = 1.0
    if d0 is None:
        d0 = 1.0

    y, *_ = boxcox(y)
    y -= y.mean()

    return least_squares(_estimate, [rho, d0], args=(x, y), method='lm').x


def make_surrogate(x, y, rho=None, d0=None, seed=None, return_order=False,
                   return_params=False):
    """
    Generates surrogate map of `y`, retaining characteristic spatial features

    Parameters
    ----------
    x : array_like
        Distance matrix
    y : array_like
        Dependent brain-imaging variable; all values must be positive
    rho : float, optional
        Estimate for rho parameter. If not provided will be estimated from
        input data. Default: None
    d0 : float, optional
        Estimate for d0 parameter. If not provided will be estimated from input
        data. Default: None
    return_order : bool, optional
        Whether to return rank order of generated `surrogate` before values
        were replaced with `y`

    Returns
    -------
    surrogate : array_like
        Input `y` matrix, permuted according to surrogate map with similar
        spatial autocorrelation factor
    order : array_like
        Rank-order of `surrogate` before values were replaced with `y`
    """

    rs = np.random.default_rng(seed)

    if rho is None or d0 is None:
        rho, d0 = estimate_rho_d0(x, y, rho=rho, d0=d0)

    w = _make_weight_matrix(x, d0)
    u = rs.standard_normal(len(x))
    i = np.identity(len(x))
    surr = np.linalg.solve(i - rho * w, u)

    order = surr.argsort()
    surr[order] = np.sort(y)

    out = (surr,)

    if return_order:
        out += (order,)
    if return_params:
        out += ((rho, d0),)

    return out[0] if len(out) == 1 else out


def batch_surrogates(x, y, rho=None, d0=None, seed=None, n_surr=1000,
                     n_jobs=1):
    """
    Generates `n_surr` surrogates maps of `y` using Burt-2018 method

    Parameters
    ----------
    x : (N, N) array_like
        Distance matrix
    y : (N,) array_like
        Dependent brain-imaging variable; all values must be positive
    n_surr : int, optional
        Number of surrogates maps to generate. Default: 1000
    n_jobs : int, optional
        Number of processes to use while generating surrogate maps. Default: 1
    seed : {int, None}, optional
        Random seed for generating surrogates. Default: None

    Returns
    -------
    surrs : (N, `n_surr`)
        Generated surrogate maps
    """

    try:
        from joblib import Parallel, delayed
        joblib_avail = True
    except ImportError:
        if n_jobs != 1:
            warnings.warn('joblib not available; cannot parallelize')
        joblib_avail = False

    def _quick_surr(iw, ysort, seed=None):
        rs = np.random.default_rng(seed)
        u = rs.standard_normal(iw.shape[0])
        if ssp.issparse(iw):
            surr = ssp.linalg.spsolve(iw, u)
        else:
            surr = np.linalg.solve(iw, u)
        surr[surr.argsort()] = ysort

        return surr

    rs = np.random.default_rng(seed)
    seeds = rs.integers(np.iinfo(np.int32).max, size=n_surr)

    if rho is None or d0 is None:
        rho, d0 = estimate_rho_d0(x, y)
    iw = np.identity(len(x)) - rho * _make_weight_matrix(x, d0)
    zeros = np.isclose(iw, 0)
    if (zeros.sum() / iw.size) > 0.5:
        iw[np.isclose(iw, 0)] = 0
        iw = ssp.csr_matrix(iw)
    ysort = np.sort(y)

    if joblib_avail:
        surrs = Parallel(n_jobs=n_jobs)(
            delayed(_quick_surr)(iw, ysort, seed=seed) for seed in seeds
        )
    else:
        surrs = [_quick_surr(iw, ysort, seed=seed) for seed in seeds]

    return np.column_stack(surrs)



# -*- coding: utf-8 -*-
"""
Functions for working with CIVET data
"""

import os

import numpy as np

from neuromaps.points import get_shared_triangles, which_triangle


def read_civet_surf(fname):
    """
    Reads a CIVET-style .obj geometry file

    Parameters
    ----------
    fname : str or os.PathLike
        Filepath to .obj file

    Returns
    -------
    vertices : (N, 3)
        Vertices of surface mesh
    triangles : (T, 3)
        Triangles comprising surface mesh
    """

    k, polygons = 0, []
    with open(fname, 'r') as src:
        n_vert = int(src.readline().split()[6])
        vertices = np.zeros((n_vert, 3))
        for i, line in enumerate(src):
            if i < n_vert:
                vertices[i] = [float(i) for i in line.split()]
            elif i >= (2 * n_vert) + 5:
                if not line.strip():
                    k = 1
                elif k == 1:
                    polygons.extend([int(i) for i in line.split()])

    triangles = np.reshape(np.asarray(polygons), (-1, 3))

    return vertices, triangles


def read_surfmap(surfmap):
    """
    Reads surface map from CIVET

    Parameters
    ----------
    surfmap : str or os.PathLike
        Surface mapping file to be loaded

    Returns
    -------
    control : (N,) array_like
        Control vertex IDs
    v0, v1 : (N,) array_like
        Target vertex IDs
    t : (N, 3) array_like
        Resampling weights
    """

    control, v0, v1, t1, t2 = np.loadtxt(surfmap, skiprows=4).T
    control = control.astype(int)
    v0 = v0.astype(int)
    v1 = v1.astype(int)
    t0 = 1 - t1 - t2

    return control, v0, v1, np.column_stack((t0, t1, t2))


def resample_surface_map(source, morph, target, surfmap):
    """
    Resamples `morph` data defined on `source` surface to `target` surface

    Uses `surfmap` to define mapping

    Inputs
    ------
    source : str or os.PathLike
        Path to surface file on which `morph` is defined
    morph : str or os.PathLike
        Path to morphology data defined on `source` surface
    target : str or os.PathLike
        Path to surface file on which to resample `morph` data
    surfmap : str or os.PathLike
        Path to surface mapping file defining transformation (CIVET style)

    Returns
    -------
    resampled : np.ndarray
        Provided `morph` data resampled to `target` surface
    """

    if isinstance(source, (str, os.PathLike)):
        source = read_civet_surf(source)
    if isinstance(morph, (str, os.PathLike)):
        morph = np.loadtxt(morph)
    if len(morph) != len(source[0]):
        raise ValueError('Provided `morph` file has different number of '
                         'vertices from provided `source` surface')

    if isinstance(target, (str, os.PathLike)):
        target = read_civet_surf(target)
    if isinstance(surfmap, (str, os.PathLike)):
        surfmap = read_surfmap(surfmap)
    if len(surfmap[0]) != len(target[0]):
        raise ValueError('Provided `target` surface has different number of '
                         'vertices from provided `surfmap` transformation.')

    source_tris = get_shared_triangles(source[1])
    resampled = np.zeros_like(morph)
    for (control, v0, v1, t) in zip(*surfmap):
        tris = source_tris[(v0, v1) if v0 < v1 else (v1, v0)]
        point, verts = target[0][control], source[0][tris]
        idx = which_triangle(point, verts)
        if idx is None:
            idx = np.argmin(np.linalg.norm(point - verts[:, -1], axis=1))
        resampled[control] = np.sum(morph[[v0, v1, tris[idx][-1]]] * t)

    return resampled



# -*- coding: utf-8 -*-
"""
For testing neuromaps.datasets.utils functionality
"""

import os
import pytest

from neuromaps.datasets import utils


@pytest.mark.xfail
def test__osfify_urls():
    assert False


@pytest.mark.xfail
def test_get_dataset_info():
    assert False


@pytest.mark.xfail
def test_get_data_dir():
    assert False


def test__get_token():
    orig = os.environ.pop('NEUROMAPS_OSF_TOKEN', None)
    assert utils._get_token(None) is None
    assert utils._get_token('test') == 'test'
    os.environ['NEUROMAPS_OSF_TOKEN'] = 'test_env'
    assert utils._get_token(None) == 'test_env'
    assert utils._get_token('test') == 'test'
    if orig is not None:  # reset env variable
        os.environ['NEUROMAPS_OSF_TOKEN'] = orig


@pytest.mark.xfail
def test__get_session():
    assert False



"""
Functions for fetching datasets
"""

__all__ = [
    'fetch_all_atlases', 'fetch_atlas', 'fetch_civet', 'fetch_fsaverage',
    'fetch_fslr', 'fetch_mni152', 'fetch_regfusion', 'get_atlas_dir',
    'DENSITIES', 'ALIAS', 'available_annotations', 'available_tags',
    'fetch_annotation'
]

# TODO: remove after nilearn v0.9 release
import warnings
warnings.filterwarnings('ignore', message='Fetchers from the nilearn.datasets',
                        category=FutureWarning)

from .atlases import (fetch_all_atlases, fetch_atlas, fetch_civet,  # noqa
                      fetch_fsaverage, fetch_fslr, fetch_mni152,
                      fetch_regfusion, get_atlas_dir, DENSITIES, ALIAS)
from .annotations import (available_annotations, available_tags,  # noqa
                          fetch_annotation)
from .contributions import (upload_annotation)